# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import types
from typing import Any
from unittest.mock import MagicMock

import pytest

from anemoi.models.data_indices.collection import IndexCollection
from anemoi.training.diagnostics.callbacks.sanity import CheckVariableOrder
from anemoi.training.train.tasks.forecaster import GraphForecaster
from anemoi.training.train.train import AnemoiTrainer


@pytest.fixture
def name_to_index() -> dict:
    return {"a": 0, "b": 1, "c": 2}


@pytest.fixture
def name_to_index_permute() -> dict:
    return {"a": 0, "b": 2, "c": 1}


@pytest.fixture
def name_to_index_rename() -> dict:
    return {"a": 0, "b": 1, "d": 2}


@pytest.fixture
def name_to_index_partial_rename_permute() -> dict:
    return {"a": 2, "b": 1, "d": 0}


@pytest.fixture
def name_to_index_rename_permute() -> dict:
    return {"x": 2, "b": 1, "d": 0}


@pytest.fixture
def fake_trainer(mocker: Any, name_to_index: dict) -> AnemoiTrainer:
    trainer = mocker.Mock(spec=AnemoiTrainer)
    trainer.datamodule.data_indices.name_to_index = name_to_index
    trainer.datamodule.data_indices.compare_variables = types.MethodType(
        IndexCollection.compare_variables,
        trainer.datamodule.data_indices,
    )
    return trainer


@pytest.fixture
def fake_pl_module(mocker: Any, name_to_index: dict) -> MagicMock:
    pl_module = mocker.Mock()
    pl_module._ckpt_model_name_to_index = name_to_index
    return pl_module


@pytest.fixture
def callback() -> CheckVariableOrder:
    callback = CheckVariableOrder()
    assert callback is not None
    assert hasattr(callback, "on_train_start")
    assert hasattr(callback, "on_validation_start")
    assert hasattr(callback, "on_test_start")

    return callback


def test_on_epoch(
    fake_trainer: AnemoiTrainer,
    fake_pl_module: MagicMock,
    callback: CheckVariableOrder,
    name_to_index: dict,
) -> None:
    """Test all epoch functions with "working" indices."""
    fake_trainer.datamodule.ds_train.name_to_index = name_to_index
    fake_trainer.datamodule.ds_valid.name_to_index = name_to_index
    fake_trainer.datamodule.ds_test.name_to_index = name_to_index
    callback.on_train_start(fake_trainer, fake_pl_module)
    callback.on_validation_start(fake_trainer, fake_pl_module)
    callback.on_test_start(fake_trainer, fake_pl_module)

    assert (
        fake_trainer.datamodule.data_indices.compare_variables(
            fake_pl_module._ckpt_model_name_to_index,
            name_to_index,
        )
        is None
    )


def test_on_epoch_permute(
    fake_trainer: AnemoiTrainer,
    fake_pl_module: MagicMock,
    callback: CheckVariableOrder,
    name_to_index_permute: dict,
) -> None:
    """Test all epoch functions with permuted indices.

    Expecting errors in all cases.
    """
    fake_trainer.datamodule.ds_train.name_to_index = name_to_index_permute
    fake_trainer.datamodule.ds_valid.name_to_index = name_to_index_permute
    fake_trainer.datamodule.ds_test.name_to_index = name_to_index_permute
    with pytest.raises(ValueError, match="Detected a different sort order of the same variables:") as exc_info:
        callback.on_train_start(fake_trainer, fake_pl_module)
    assert "{'c': (2, 1), 'b': (1, 2)}" in str(exc_info.value) or "{'b': (1, 2), 'c': (2, 1)}" in str(exc_info.value)
    with pytest.raises(ValueError, match="Detected a different sort order of the same variables:") as exc_info:
        callback.on_validation_start(fake_trainer, fake_pl_module)
    assert "{'c': (2, 1), 'b': (1, 2)}" in str(exc_info.value) or "{'b': (1, 2), 'c': (2, 1)}" in str(exc_info.value)
    with pytest.raises(ValueError, match="Detected a different sort order of the same variables:") as exc_info:
        callback.on_test_start(fake_trainer, fake_pl_module)
    assert "{'c': (2, 1), 'b': (1, 2)}" in str(exc_info.value) or "{'b': (1, 2), 'c': (2, 1)}" in str(exc_info.value)

    with pytest.raises(ValueError, match="Detected a different sort order of the same variables:") as exc_info:
        fake_trainer.datamodule.data_indices.compare_variables(
            fake_pl_module._ckpt_model_name_to_index,
            name_to_index_permute,
        )
    assert "{'c': (2, 1), 'b': (1, 2)}" in str(exc_info.value) or "{'b': (1, 2), 'c': (2, 1)}" in str(exc_info.value)


def test_on_epoch_rename(
    fake_trainer: AnemoiTrainer,
    fake_pl_module: MagicMock,
    callback: CheckVariableOrder,
    name_to_index_rename: dict,
) -> None:
    """Test all epoch functions with renamed indices.

    Expecting passes in all cases.
    """
    fake_trainer.datamodule.ds_train.name_to_index = name_to_index_rename
    fake_trainer.datamodule.ds_valid.name_to_index = name_to_index_rename
    fake_trainer.datamodule.ds_test.name_to_index = name_to_index_rename
    callback.on_train_start(fake_trainer, fake_pl_module)
    callback.on_validation_start(fake_trainer, fake_pl_module)
    callback.on_test_start(fake_trainer, fake_pl_module)

    fake_trainer.datamodule.data_indices.compare_variables(
        fake_pl_module._ckpt_model_name_to_index,
        name_to_index_rename,
    )


def test_on_epoch_rename_permute(
    fake_trainer: AnemoiTrainer,
    fake_pl_module: MagicMock,
    callback: CheckVariableOrder,
    name_to_index_rename_permute: dict,
) -> None:
    """Test all epoch functions with renamed and permuted indices.

    Expects all passes (but warnings).
    """
    fake_trainer.datamodule.ds_train.name_to_index = name_to_index_rename_permute
    fake_trainer.datamodule.ds_valid.name_to_index = name_to_index_rename_permute
    fake_trainer.datamodule.ds_test.name_to_index = name_to_index_rename_permute
    callback.on_train_start(fake_trainer, fake_pl_module)
    callback.on_validation_start(fake_trainer, fake_pl_module)
    callback.on_test_start(fake_trainer, fake_pl_module)

    fake_trainer.datamodule.data_indices.compare_variables(
        fake_pl_module._ckpt_model_name_to_index,
        name_to_index_rename_permute,
    )


def test_on_epoch_partial_rename_permute(
    fake_trainer: AnemoiTrainer,
    fake_pl_module: MagicMock,
    callback: CheckVariableOrder,
    name_to_index_partial_rename_permute: dict,
) -> None:
    """Test all epoch functions with partially renamed and permuted indices.

    Expects all errors.
    """
    fake_trainer.datamodule.ds_train.name_to_index = name_to_index_partial_rename_permute
    fake_trainer.datamodule.ds_valid.name_to_index = name_to_index_partial_rename_permute
    fake_trainer.datamodule.ds_test.name_to_index = name_to_index_partial_rename_permute
    with pytest.raises(ValueError, match=r"The variable order in the model and data is different."):
        callback.on_train_start(fake_trainer, fake_pl_module)
    with pytest.raises(ValueError, match=r"The variable order in the model and data is different."):
        callback.on_validation_start(fake_trainer, fake_pl_module)
    with pytest.raises(ValueError, match=r"The variable order in the model and data is different."):
        callback.on_test_start(fake_trainer, fake_pl_module)

    with pytest.raises(ValueError, match=r"The variable order in the model and data is different."):
        fake_trainer.datamodule.data_indices.compare_variables(
            fake_pl_module._ckpt_model_name_to_index,
            name_to_index_partial_rename_permute,
        )


def test_on_epoch_wrong_validation(
    fake_trainer: AnemoiTrainer,
    fake_pl_module: MagicMock,
    callback: CheckVariableOrder,
    name_to_index: dict,
    name_to_index_permute: dict,
    name_to_index_rename: dict,
) -> None:
    """Test all epoch functions with "working" indices, but different validation indices."""
    fake_trainer.datamodule.ds_train.name_to_index = name_to_index
    fake_trainer.datamodule.ds_valid.name_to_index = name_to_index_permute
    fake_trainer.datamodule.ds_test.name_to_index = name_to_index_rename
    callback.on_train_start(fake_trainer, fake_pl_module)
    with pytest.raises(ValueError, match="Detected a different sort order of the same variables:") as exc_info:
        callback.on_validation_start(fake_trainer, fake_pl_module)
    assert " {'c': (2, 1), 'b': (1, 2)}" in str(
        exc_info.value,
    ) or "{'b': (1, 2), 'c': (2, 1)}" in str(exc_info.value)
    callback.on_test_start(fake_trainer, fake_pl_module)

    assert (
        fake_trainer.datamodule.data_indices.compare_variables(
            fake_pl_module._ckpt_model_name_to_index,
            name_to_index,
        )
        is None
    )


def test_on_load_checkpoint_restores_name_to_index() -> None:

    model = GraphForecaster.__new__(GraphForecaster)

    model.on_load_checkpoint = types.MethodType(GraphForecaster.on_load_checkpoint, GraphForecaster)

    mock_name_to_index = {"var1": 0, "var2": 1}
    mock_checkpoint = {"hyper_parameters": {"data_indices": MagicMock(name_to_index=mock_name_to_index)}}
    # Act
    model.on_load_checkpoint(mock_checkpoint)

    # Assert
    assert model._ckpt_model_name_to_index == mock_name_to_index



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import os


def get_base_seed(base_seed_env: str | None = None) -> int:
    """Gets the base seed from the environment variables.

    Option to manually set a seed via export ANEMOI_BASE_SEED=xxx in job script

    Parameters
    ----------
    base_seed_env : str, optional
        Environment variable to use for the base seed, by default None

    Returns
    -------
    int
        Base seed.

    """
    env_var_list = ["ANEMOI_BASE_SEED", "SLURM_JOB_ID"]
    if base_seed_env is not None:
        env_var_list = [base_seed_env, *env_var_list]

    base_seed = None
    for env_var in env_var_list:
        if env_var in os.environ:
            base_seed = int(os.environ.get(env_var))
            break

    assert base_seed is not None, f"Base seed not found in environment variables {env_var_list}"

    base_seed_threshold = 1000
    if base_seed < base_seed_threshold:
        base_seed *= base_seed_threshold  # make it (hopefully) big enough

    return base_seed



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from torch.utils.checkpoint import checkpoint

from anemoi.training.train.tasks.rollout import BaseRolloutGraphModule

if TYPE_CHECKING:
    from collections.abc import Generator

    import torch


LOGGER = logging.getLogger(__name__)


class GraphForecaster(BaseRolloutGraphModule):
    """Graph neural network forecaster for PyTorch Lightning."""

    def _rollout_step(
        self,
        batch: torch.Tensor,
        rollout: int | None = None,
        validation_mode: bool = False,
    ) -> Generator[tuple[torch.Tensor | None, dict, list]]:
        """Rollout step for the forecaster.

        Parameters
        ----------
        batch : torch.Tensor
            Normalized batch to use for rollout (assumed to be already preprocessed)
        rollout : Optional[int], optional
            Number of times to rollout for, by default None
            If None, will use self.rollout
        validation_mode : bool, optional
            Whether in validation mode, and to calculate validation metrics, by default False
            If False, metrics will be empty

        Yields
        ------
        Generator[tuple[Union[torch.Tensor, None], dict, list], None, None]
            Loss value, metrics, and predictions (per step)

        """
        # start rollout of preprocessed batch
        x = batch[
            :,
            0 : self.multi_step,
            ...,
            self.data_indices.data.input.full,
        ]  # (bs, multi_step, latlon, nvar)
        msg = (
            "Batch length not sufficient for requested multi_step length!"
            f", {batch.shape[1]} !>= {rollout + self.multi_step}"
        )
        assert batch.shape[1] >= rollout + self.multi_step, msg

        for rollout_step in range(rollout or self.rollout):
            # prediction at rollout step rollout_step, shape = (bs, latlon, nvar)
            y_pred = self(x)

            y = batch[:, self.multi_step + rollout_step, ..., self.data_indices.data.output.full]
            LOGGER.debug("SHAPE: y.shape = %s", list(y.shape))
            # y includes the auxiliary variables, so we must leave those out when computing the loss

            loss, metrics_next, y_pred = checkpoint(
                self.compute_loss_metrics,
                y_pred,
                y,
                step=rollout_step,
                validation_mode=validation_mode,
                use_reentrant=False,
            )

            x = self._advance_input(x, y_pred, batch, rollout_step)

            yield loss, metrics_next, y_pred



# (C) Copyright 2024-2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from enum import Enum
from functools import partial
from typing import Annotated
from typing import Any
from typing import Literal

from pydantic import AfterValidator
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Discriminator
from pydantic import Field
from pydantic import NonNegativeFloat
from pydantic import NonNegativeInt
from pydantic import PositiveInt
from pydantic import field_validator
from pydantic import model_validator
from typing_extensions import Self

from anemoi.utils.schemas import BaseModel
from anemoi.utils.schemas.errors import allowed_values


class GradientClip(BaseModel):
    """Gradient clipping configuration."""

    val: float = 32.0
    "Gradient clipping value."
    algorithm: Annotated[str, AfterValidator(partial(allowed_values, values=["value", "norm"]))] = Field(
        example="value",
    )
    "The gradient clipping algorithm to use"


class SWA(BaseModel):
    """Stochastic weight averaging configuration.

    See https://pytorch.org/blog/stochastic-weight-averaging-in-pytorch/
    """

    enabled: bool = Field(example=False)
    "Enable stochastic weight averaging."
    lr: NonNegativeFloat = Field(example=1.0e-4)
    "Learning rate for SWA."


class Rollout(BaseModel):
    """Rollout configuration."""

    start: PositiveInt = Field(example=1)
    "Number of rollouts to start with."
    epoch_increment: NonNegativeInt = Field(example=0)
    "Number of epochs to increment the rollout."
    max: PositiveInt = Field(example=1)
    "Maximum number of rollouts."


class LR(BaseModel):
    """Learning rate configuration.

    Changes in per-gpu batch_size should come with a rescaling of the local_lr,
    in order to keep a constant global_lr global_lr = local_lr * num_gpus_per_node * num_nodes / gpus_per_model.
    """

    rate: NonNegativeFloat = Field(example=0.625e-4)  # TODO(Helen): Could be computed by pydantic
    "Initial learning rate. Is adjusteed according to the hardware configuration"
    iterations: NonNegativeInt = Field(example=300000)
    "Number of iterations."
    min: NonNegativeFloat = Field(example=3e-7)
    "Minimum learning rate."
    warmup: NonNegativeInt = Field(example=1000)
    "Number of warm up iteration. Default to 1000."


class OptimizerSchema(PydanticBaseModel):
    """Choosing the PydanticBaseModel to allow extra inputs."""

    target_: str = Field(..., alias="_target_")
    """Full path to the optimizer class, e.g. `torch.optim.AdamW`."""


class ExplicitTimes(BaseModel):
    """Time indices for input and output.

    Starts at index 0. Input and output can overlap.
    """

    input: list[NonNegativeInt] = Field(examples=[0, 1])
    "Input time indices."
    target: list[NonNegativeInt] = Field(examples=[2])
    "Target time indices."


class TargetForcing(BaseModel):
    """Forcing parameters for target output times.

    Extra forcing parameters to use as input to distinguish between different target times.
    """

    data: list[str] = Field(examples=["insolation"])
    "List of forcing parameters to use as input to the model at the interpolated step."
    time_fraction: bool = Field(example=True)
    "Use target time as a fraction between input boundary times as input."


class LossScalingSchema(BaseModel):
    default: int = 1
    "Default scaling value applied to the variables loss. Default to 1."
    pl: dict[str, NonNegativeFloat]
    "Scaling value associated to each pressure level variable loss."
    sfc: dict[str, NonNegativeFloat]
    "Scaling value associated to each surface variable loss."


class GeneralVariableLossScalerSchema(BaseModel):
    target_: Literal["anemoi.training.losses.scalers.GeneralVariableLossScaler"] = Field(..., alias="_target_")
    weights: dict[str, float]
    "Weight of each variable."  # Check keys (variables) are read ???


class VariableMaskingScalerSchema(BaseModel):
    target_: Literal["anemoi.training.losses.scalers.VariableMaskingLossScaler"] = Field(..., alias="_target_")
    variables: list[str] = Field(defaultexample=["tp"])
    "Variables to compute the loss over."
    invert: bool = Field(examples=False)
    "Flag to invert the variable mask."


class NaNMaskScalerSchema(BaseModel):
    target_: Literal["anemoi.training.losses.scalers.NaNMaskScaler"] = Field(..., alias="_target_")
    use_processors_tendencies: bool = Field(default=False)
    "Flag to include processors for tendencies when building the loss mask."


class TendencyScalerTargets(str, Enum):
    stdev = "anemoi.training.losses.scalers.StdevTendencyScaler"
    var = "anemoi.training.losses.scalers.VarTendencyScaler"


class TendencyScalerSchema(BaseModel):
    target_: TendencyScalerTargets = Field(
        example="anemoi.training.losses.scalers.StdevTendencyScaler",
        alias="_target_",
    )


class VariableLevelScalerTargets(str, Enum):
    relu_scaler = "anemoi.training.losses.scalers.ReluVariableLevelScaler"
    linear_scaler = "anemoi.training.losses.scalers.LinearVariableLevelScaler"
    polynomial_sclaer = "anemoi.training.losses.scalers.PolynomialVariableLevelScaler"
    no_scaler = "anemoi.training.losses.scalers.NoVariableLevelScaler"


class VariableLevelScalerSchema(BaseModel):
    target_: VariableLevelScalerTargets = Field(
        example="anemoi.training.losses.scalers.ReluVariableLevelScaler",
        alias="_target_",
    )
    group: str = Field(example="pl")
    "Group of variables to scale."
    slope: float = Field(example=1.0)
    "Slope of scaling function."
    y_intercept: float = Field(example=0.001)
    "Y-axis shift of scaling function."


class GraphNodeAttributeScalerSchema(BaseModel):
    target_: Literal["anemoi.training.losses.scalers.GraphNodeAttributeScaler"] = Field(..., alias="_target_")
    nodes_name: str = Field(example="data")
    "Name of the nodes to take the attribute from."
    nodes_attribute_name: str = Field(example="area_weight")
    "Name of the node attribute to return."
    norm: Literal["unit-max", "unit-sum"] | None = Field(example="unit-sum")
    "Normalisation method applied to the node attribute."


class ReweightedGraphNodeAttributeScalerSchema(BaseModel):
    target_: Literal["anemoi.training.losses.scalers.ReweightedGraphNodeAttributeScaler"] = Field(
        ...,
        alias="_target_",
    )
    nodes_name: str = Field(example="data")
    "Name of the nodes to take the attribute from."
    nodes_attribute_name: str = Field(example="area_weight")
    "Name of the node attribute to return."
    scaling_mask_attribute_name: str = Field(example="cutout_mask")
    "Name of the node attribute to use as a mask to reweight the reference values."
    weight_frac_of_total: float = Field(example=0.5)
    "Fraction of total weight to assign to nodes within the scaling mask. The remaining weight is distributed among "
    "nodes outside the mask."
    norm: Literal["unit-max", "unit-sum"] | None = Field(example="unit-sum")
    "Normalisation method applied to the node attribute."


ScalerSchema = (
    GeneralVariableLossScalerSchema
    | VariableLevelScalerSchema
    | VariableMaskingScalerSchema
    | TendencyScalerSchema
    | NaNMaskScalerSchema
    | GraphNodeAttributeScalerSchema
    | ReweightedGraphNodeAttributeScalerSchema
)


class ImplementedLossesUsingBaseLossSchema(str, Enum):
    kcrps = "anemoi.training.losses.kcrps.KernelCRPS"
    afkcrps = "anemoi.training.losses.kcrps.AlmostFairKernelCRPS"
    rmse = "anemoi.training.losses.RMSELoss"
    mse = "anemoi.training.losses.MSELoss"
    weighted_mse = "anemoi.training.losses.WeightedMSELoss"
    mae = "anemoi.training.losses.MAELoss"
    logcosh = "anemoi.training.losses.LogCoshLoss"
    huber = "anemoi.training.losses.HuberLoss"
    combined = "anemoi.training.losses.combined.CombinedLoss"


class BaseLossSchema(BaseModel):
    target_: ImplementedLossesUsingBaseLossSchema = Field(..., alias="_target_")
    "Loss function object from anemoi.training.losses."
    scalers: list[str] = Field(example=["variable"])  # TODO(Mario): Validate scalers are defined
    "Scalars to include in loss calculation"
    ignore_nans: bool = False
    "Allow nans in the loss and apply methods ignoring nans for measuring the loss."


class KernelCRPSSchema(BaseLossSchema):
    fair: bool = True
    "Calculate a 'fair' (unbiased) score - ensemble variance component weighted by (ens-size-1)^-1"


class AlmostFairKernelCRPSSchema(BaseLossSchema):
    alpha: float = 1.0
    """Factor for linear combination of fair (unbiased, ensemble variance component
    weighted by (ens-size-1)^-1) and standard CRPS (1.0 = fully fair, 0.0 = fully unfair)"""
    no_autocast: bool = True
    "Deactivate autocast for the kernel CRPS calculation"


class MultiScaleLossSchema(BaseModel):
    target_: Literal["anemoi.training.losses.MultiscaleLossWrapper"] = Field(..., alias="_target_")
    per_scale_loss: AlmostFairKernelCRPSSchema | KernelCRPSSchema
    weights: list[float]
    keep_batch_sharded: bool
    loss_matrices_path: str
    loss_matrices: list[str | None]

    @field_validator("weights")
    @classmethod
    def validate_weights_length(cls, v: list[float], info: Any) -> list[float]:
        if "loss_matrices" in info.data:
            assert len(v) == len(info.data["loss_matrices"]), "weights must have same length as loss_matrices"
        return v


class HuberLossSchema(BaseLossSchema):
    delta: float = 1.0
    "Threshold for Huber loss."


class CombinedLossSchema(BaseLossSchema):
    losses: list[BaseLossSchema] = Field(min_length=1)
    "Losses to combine, can be any of the normal losses."
    loss_weights: list[int | float] | None = None
    "Weightings of losses, if not set, all losses are weighted equally."

    @field_validator("losses", mode="before")
    @classmethod
    def add_empty_scalers(cls, losses: Any) -> Any:
        """Add empty scalers to loss functions, as scalers can be set at top level."""
        from omegaconf.omegaconf import open_dict

        for loss in losses:
            if "scalers" not in loss:
                with open_dict(loss):
                    loss["scalers"] = []
        return losses

    @model_validator(mode="after")
    def check_length_of_weights_and_losses(self) -> Self:
        """Check that the number of losses and weights match, or if not set, skip."""
        losses, loss_weights = self.losses, self.loss_weights
        if loss_weights is not None and len(losses) != len(loss_weights):
            error_msg = "Number of losses and weights must match"
            raise ValueError(error_msg)
        return self


LossSchemas = (
    BaseLossSchema
    | HuberLossSchema
    | CombinedLossSchema
    | AlmostFairKernelCRPSSchema
    | KernelCRPSSchema
    | MultiScaleLossSchema
)


class ImplementedStrategiesUsingBaseDDPStrategySchema(str, Enum):
    ddp_ens = "anemoi.training.distributed.strategy.DDPEnsGroupStrategy"
    ddp = "anemoi.training.distributed.strategy.DDPGroupStrategy"


class BaseDDPStrategySchema(BaseModel):
    """Strategy configuration."""

    target_: ImplementedStrategiesUsingBaseDDPStrategySchema = Field(..., alias="_target_")
    num_gpus_per_model: PositiveInt = Field(example=2)
    "Number of GPUs per model."
    read_group_size: PositiveInt = Field(example=1)
    "Number of GPUs per reader group. Defaults to number of GPUs."


class DDPEnsGroupStrategyStrategySchema(BaseDDPStrategySchema):
    """Strategy object from anemoi.training.strategy."""

    num_gpus_per_ensemble: PositiveInt = Field(example=2)
    "Number of GPUs per ensemble."


StrategySchemas = BaseDDPStrategySchema | DDPEnsGroupStrategyStrategySchema


class BaseTrainingSchema(BaseModel):
    """Training configuration."""

    run_id: str | None = Field(example=None)
    "Run ID: used to resume a run from a checkpoint, either last.ckpt or specified in system.input.warm_start."
    fork_run_id: str | None = Field(example=None)
    "Run ID to fork from, either last.ckpt or specified in system.input.warm_start."
    load_weights_only: bool = Field(example=False)
    "Load only the weights from the checkpoint, not the optimiser state."
    transfer_learning: bool = Field(example=False)
    "Flag to activate transfer learning mode when loading a checkpoint."
    submodules_to_freeze: list[str] = Field(example=["processor"])
    "List of submodules to freeze during transfer learning."
    deterministic: bool = Field(default=False)
    "This flag sets the torch.backends.cudnn.deterministic flag. Might be slower, but ensures reproducibility."
    precision: str = Field(default="16-mixed")
    "Precision"
    multistep_input: PositiveInt = Field(example=2)
    """Number of input steps for the model. E.g. 1 = single step scheme, X(t-1) used to predict X(t),
    k > 1: multistep scheme, uses [X(t-k), X(t-k+1), ... X(t-1)] to predict X(t)."""
    accum_grad_batches: PositiveInt = Field(default=1)
    """Accumulates gradients over k batches before stepping the optimizer.
    K >= 1 (if K == 1 then no accumulation). The effective bacthsize becomes num-device * k."""
    num_sanity_val_steps: NonNegativeInt = Field(example=6)
    "Sanity check runs n batches of val before starting the training routine."
    gradient_clip: GradientClip
    "Config for gradient clipping."
    strategy: StrategySchemas
    "Strategy to use."
    swa: SWA = Field(default_factory=SWA)
    "Config for stochastic weight averaging."
    training_loss: LossSchemas
    "Training loss configuration."
    loss_gradient_scaling: bool = False
    "Dynamic rescaling of the loss gradient. Not yet tested."
    scalers: dict[str, ScalerSchema]
    "Scalers to use in the computation of the loss and validation scores."
    validation_metrics: dict[str, LossSchemas]
    "List of validation metrics configurations."
    variable_groups: dict[str, str | list[str] | dict[str, str | bool | list[str | int]]]
    "Groups for variable loss scaling"
    max_epochs: PositiveInt | None = None
    "Maximum number of epochs, stops earlier if max_steps is reached first."
    max_steps: PositiveInt = 150000
    "Maximum number of steps, stops earlier if max_epochs is reached first."
    lr: LR = Field(default_factory=LR)
    "Learning rate configuration."
    optimizer: OptimizerSchema = Field(default_factory=OptimizerSchema)
    "Optimizer configuration."
    recompile_limit: PositiveInt = 32
    "How many times torch.compile will recompile a function for a given input shape."
    metrics: list[str]
    "List of metrics"
    ensemble_size_per_device: PositiveInt = 1
    "Number of ensemble members per device. Default is 1 for non-ensemble forecasting."


class ForecasterSchema(BaseTrainingSchema):
    model_task: Literal["anemoi.training.train.tasks.GraphForecaster",] = Field(..., alias="model_task")
    "Training objective."
    rollout: Rollout = Field(default_factory=Rollout)
    "Rollout configuration."


class ForecasterEnsSchema(ForecasterSchema):
    model_task: Literal["anemoi.training.train.tasks.GraphEnsForecaster",] = Field(..., alias="model_task")
    "Training objective."


class DiffusionForecasterSchema(ForecasterSchema):
    model_task: Literal["anemoi.training.train.tasks.GraphDiffusionForecaster"] = Field(..., alias="model_task")
    "Training objective."


class DiffusionTendForecasterSchema(ForecasterSchema):
    model_task: Literal["anemoi.training.train.tasks.GraphDiffusionTendForecaster"] = Field(
        ...,
        alias="model_task",
    )
    "Training objective."


class InterpolationSchema(BaseTrainingSchema):
    model_task: Literal["anemoi.training.train.tasks.GraphInterpolator"] = Field(..., alias="model_task")
    "Training objective."
    explicit_times: ExplicitTimes
    "Time indices for input and output."
    target_forcing: TargetForcing
    "Forcing parameters for target output times."


TrainingSchema = Annotated[
    ForecasterSchema
    | ForecasterEnsSchema
    | InterpolationSchema
    | DiffusionForecasterSchema
    | DiffusionTendForecasterSchema,
    Discriminator("model_task"),
]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from omegaconf import DictConfig
from pytorch_lightning.callbacks import LearningRateMonitor as pl_LearningRateMonitor
from pytorch_lightning.callbacks.stochastic_weight_avg import StochasticWeightAveraging as pl_StochasticWeightAveraging

LOGGER = logging.getLogger(__name__)


class LearningRateMonitor(pl_LearningRateMonitor):
    """Provide LearningRateMonitor from pytorch_lightning as a callback."""

    def __init__(
        self,
        config: DictConfig,
        logging_interval: str = "step",
        log_momentum: bool = False,
    ) -> None:
        super().__init__(logging_interval=logging_interval, log_momentum=log_momentum)
        self.config = config


class StochasticWeightAveraging(pl_StochasticWeightAveraging):
    """Provide StochasticWeightAveraging from pytorch_lightning as a callback."""

    def __init__(
        self,
        config: DictConfig,
        swa_lrs: int | None = None,
        swa_epoch_start: int | None = None,
        annealing_epochs: int | None = None,
        annealing_strategy: str | None = None,
        device: str | None = None,
        **kwargs,
    ) -> None:
        """Stochastic Weight Averaging Callback.

        Parameters
        ----------
        config : OmegaConf
            Full configuration object
        swa_lrs : int, optional
            Stochastic Weight Averaging Learning Rate, by default None
        swa_epoch_start : int, optional
            Epoch start, by default 0.75 * config.training.max_epochs
        annealing_epochs : int, optional
            Annealing Epoch, by default 0.25 * config.training.max_epochs
        annealing_strategy : str, optional
            Annealing Strategy, by default 'cos'
        device : str, optional
            Device to use, by default None
        """
        kwargs["swa_lrs"] = swa_lrs or config.training.swa.lr
        kwargs["swa_epoch_start"] = swa_epoch_start or min(
            int(0.75 * config.training.max_epochs),
            config.training.max_epochs - 1,
        )
        kwargs["annealing_epochs"] = annealing_epochs or max(int(0.25 * config.training.max_epochs), 1)
        kwargs["annealing_strategy"] = annealing_strategy or "cos"
        kwargs["device"] = device

        super().__init__(**kwargs)
        self.config = config



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import pytorch_lightning as pl
from hydra.utils import instantiate
from omegaconf import DictConfig
from omegaconf import OmegaConf

from anemoi.training.schemas.base_schema import BaseSchema
from anemoi.training.schemas.base_schema import convert_to_omegaconf

LOGGER = logging.getLogger(__name__)


def get_mlflow_logger(config: BaseSchema) -> None:
    if not config.diagnostics.log.mlflow.enabled:
        LOGGER.debug("MLFlow logging is disabled.")
        return None

    logger_config = OmegaConf.to_container(convert_to_omegaconf(config).diagnostics.log.mlflow)
    del logger_config["enabled"]

    # backward compatibility to not break configs
    logger_config["_target_"] = logger_config.get(
        "_target_",
        "anemoi.training.diagnostics.mlflow.logger.AnemoiMLflowLogger",
    )
    logger_config["save_dir"] = logger_config.get("save_dir", str(config.system.output.logs.mlflow))

    logger = instantiate(
        logger_config,
        run_id=config.training.run_id,
        fork_run_id=config.training.fork_run_id,
    )

    if logger.log_terminal:
        logger.log_terminal_output(artifact_save_dir=config.system.output.plots)
    if logger.log_system:
        logger.log_system_metrics()

    return logger


def get_tensorboard_logger(config: DictConfig) -> pl.loggers.TensorBoardLogger | None:
    """Setup TensorBoard experiment logger.

    Parameters
    ----------
    config : DictConfig
        Job configuration

    Returns
    -------
    pl.loggers.TensorBoardLogger | None
        Logger object, or None

    """
    if not config.diagnostics.log.tensorboard.enabled:
        LOGGER.debug("Tensorboard logging is disabled.")
        return None

    from pytorch_lightning.loggers import TensorBoardLogger

    return TensorBoardLogger(
        save_dir=config.system.output.logs.tensorboard,
        log_graph=False,
    )


def get_wandb_logger(config: DictConfig, model: pl.LightningModule) -> pl.loggers.WandbLogger | None:
    """Setup Weights & Biases experiment logger.

    Parameters
    ----------
    config : DictConfig
        Job configuration
    model: GraphForecaster
        Model to watch

    Returns
    -------
    pl.loggers.WandbLogger | None
        Logger object

    Raises
    ------
    ImportError
        If `wandb` is not installed

    """
    if not config.diagnostics.log.wandb.enabled:
        LOGGER.debug("Weights & Biases logging is disabled.")
        return None

    try:
        from pytorch_lightning.loggers.wandb import WandbLogger
    except ImportError as err:
        msg = "To activate W&B logging, please install `wandb` as an optional dependency."
        raise ImportError(msg) from err

    logger = WandbLogger(
        project=config.diagnostics.log.wandb.project,
        entity=config.diagnostics.log.wandb.entity,
        id=config.training.run_id,
        save_dir=config.system.output.logs.wandb,
        offline=config.diagnostics.log.wandb.offline,
        log_model=config.diagnostics.log.wandb.log_model,
        resume=config.training.run_id is not None,
    )
    logger.log_hyperparams(OmegaConf.to_container(config, resolve=True))
    if config.diagnostics.log.wandb.gradients or config.diagnostics.log.wandb.parameters:
        if config.diagnostics.log.wandb.gradients and config.diagnostics.log.wandb.parameters:
            log_ = "all"
        elif config.diagnostics.log.wandb.gradients:
            log_ = "gradients"
        else:
            log_ = "parameters"
        logger.watch(model, log=log_, log_freq=config.diagnostics.log.interval, log_graph=False)

    return logger



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from torch.utils.data import get_worker_info

LOGGER = logging.getLogger(__name__)


def worker_init_func(worker_id: int) -> None:
    """Configures each dataset worker process.

    Calls WeatherBenchDataset.per_worker_init() on each dataset object.

    Parameters
    ----------
    worker_id : int
        Worker ID

    Raises
    ------
    RuntimeError
        If worker_info is None

    """
    worker_info = get_worker_info()  # information specific to each worker process
    if worker_info is None:
        LOGGER.error("worker_info is None! Set num_workers > 0 in your dataloader!")
        raise RuntimeError
    dataset_obj = worker_info.dataset  # the copy of the dataset held by this worker process.
    dataset_obj.per_worker_init(
        n_workers=worker_info.num_workers,
        worker_id=worker_id,
    )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import torch
import torch.fft

from anemoi.training.losses.base import FunctionalLoss
from anemoi.training.utils.enums import TensorDim

LOGGER = logging.getLogger(__name__)


def amplitude(spectrum: torch.Tensor) -> torch.Tensor:
    return torch.sqrt(spectrum.real**2 + spectrum.imag**2)


def get_power_spectra_scalar_product(
    power_spectra_real: torch.Tensor,
    power_spectra_pred: torch.Tensor,
) -> torch.Tensor:
    return power_spectra_real * power_spectra_pred.conj() + power_spectra_pred * power_spectra_real.conj()


def get_spectra(
    predicted_output: torch.Tensor,
    real_output: torch.Tensor,
    dims: tuple[int, int],
) -> tuple[torch.Tensor, torch.Tensor]:
    assert dims[0] * dims[1] == real_output.shape[TensorDim.GRID], (
        "The product of dims must match the spatial dims of the output."
        "Please use x_dim and y_dim such that field_shape=(x_dim, y_dim)."
    )
    dims_total = (*real_output.shape[: TensorDim.GRID], *dims, real_output.shape[TensorDim.VARIABLE])
    power_spectra_real = torch.fft.fft2(real_output.reshape(dims_total), dim=(-2, -3))
    power_spectra_pred = torch.fft.fft2(predicted_output.reshape(dims_total), dim=(-2, -3))
    return power_spectra_real, power_spectra_pred


def log_rfft2_distance(
    predicted_output: torch.Tensor,
    real_output: torch.Tensor,
    dims: tuple[int, int],
) -> torch.Tensor:
    r"""Calculate the log spectral distance between two fields."""
    power_spectra_real, power_spectra_pred = get_spectra(predicted_output, real_output, dims)
    epsilon = torch.finfo(real_output.dtype).eps  # Small epsilon to avoid division by zero
    power_spectra_real_sq = amplitude(power_spectra_real) ** 2
    power_spectra_pred_sq = amplitude(power_spectra_pred) ** 2
    ratio = (power_spectra_real_sq + epsilon) / (power_spectra_pred_sq + epsilon)

    def log10(x: torch.Tensor) -> torch.Tensor:
        return torch.log(x) / torch.log(torch.tensor(10.0, device=x.device, dtype=x.dtype))

    return (10 * log10(ratio)) ** 2


class LogFFT2Distance(FunctionalLoss):
    r"""The log spectral distance is used to compute the difference between spectra of two fields.

    It is also called log spectral distortion.
    When it is expressed in discrete space with L2 norm, it is defined as:
    <math>D_{LS}={\left\{ \frac{1}{N} \sum_{n=1}^N \left[ \log P(n) - \log\hat{P}(n)\right]^2\right\\}}^{1/2} ,</math>.
    All scaling and weighting is handled by the parent class.
    """

    def __init__(
        self,
        x_dim: int,
        y_dim: int,
        ignore_nans: bool = False,
    ) -> None:
        super().__init__(ignore_nans)
        LOGGER.warning(
            "LogFFT2Distance can only be used with data on 2D grids.",
        )
        self.x_dim = x_dim
        self.y_dim = y_dim

    def calculate_difference(self, pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        dist = log_rfft2_distance(pred, target, dims=(self.x_dim, self.y_dim))
        return dist.reshape(pred.shape)

    def forward(
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        squash: bool = True,
        *,
        scaler_indices: tuple[int, ...] | None = None,
        without_scalers: list[str] | list[int] | None = None,
        **kwargs,  # noqa: ARG002
    ) -> torch.Tensor:
        result = super().forward(pred, target, squash, scaler_indices=scaler_indices, without_scalers=without_scalers)
        return torch.sqrt(torch.mean(result))


class FourierCorrelationLoss(FunctionalLoss):
    r"""The log spectral distance is used to compute the difference between spectra of two fields.

    See https://arxiv.org/pdf/2410.23159.pdf for more details.
    """

    def __init__(
        self,
        x_dim: int,
        y_dim: int,
        ignore_nans: bool = False,
    ) -> None:
        super().__init__(ignore_nans)
        LOGGER.warning(
            "Fourier Correlation loss can only be used with data on 2D grids.",
        )
        self.x_dim = x_dim
        self.y_dim = y_dim

    def calculate_difference(self, pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        power_spectra_real, power_spectra_pred = get_spectra(pred, target, dims=(self.x_dim, self.y_dim))
        self.power_spectra_real = power_spectra_real
        self.power_spectra_pred = power_spectra_pred
        dist = get_power_spectra_scalar_product(power_spectra_real, power_spectra_pred).real
        return dist.reshape(pred.shape)

    def forward(
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        squash: bool = True,
        *,
        scaler_indices: tuple[int, ...] | None = None,
        without_scalers: list[str] | list[int] | None = None,
        **kwargs,  # noqa: ARG002
    ) -> torch.Tensor:
        # scaling the scaler product with node weights
        scaled_scalar_product = super().forward(
            pred,
            target,
            squash,
            scaler_indices=scaler_indices,
            without_scalers=without_scalers,
        )
        # the rest of the loss implies summing over spatial dimensions
        numerator = (1 / 2) * torch.sum(
            scaled_scalar_product,
            dim=(-2, -3),
        )
        # and scaling with the spectrum amplitude: needs to be done after scaling with node weights
        # otherwise the spatial dimension is lost
        epsilon = torch.finfo(target.dtype).eps  # Small epsilon to avoid division by zero
        denominator = torch.sqrt(
            torch.sum(amplitude(self.power_spectra_real) ** 2, dim=(-2, -3))
            * torch.sum(amplitude(self.power_spectra_pred) ** 2, dim=(-2, -3))
            + epsilon,
        )
        return torch.mean(1 - numerator / denominator)



"""AdEMAMix optimizer implementation.

-----------------------------------------------------------------------------
Copyright (c) 2024 Apple Inc.

This source file is part of the ADEMAMix project:
https://github.com/apple/ml-ademamix

Licensed under the MIT License.
See the LICENSE file at: https://github.com/apple/ml-ademamix/blob/main/LICENSE

Adapted from: https://pytorch.org/docs/1.6.0/_modules/torch/optim/adam.html
-----------------------------------------------------------------------------
"""

import math
from collections.abc import Callable
from collections.abc import Iterable
from typing import Any

import torch
from torch.optim import Optimizer


def linear_warmup_scheduler(step: int, alpha_end: float, alpha_start: float = 0.0, warmup: int = 1) -> float:
    if step < warmup:
        a = step / float(warmup)
        return (1.0 - a) * alpha_start + a * alpha_end
    return alpha_end


def linear_hl_warmup_scheduler(step: int, beta_end: float, beta_start: float = 0.0, warmup: int = 1) -> float:

    def f(beta: float, eps: float = 1e-8) -> float:
        return math.log(0.5) / math.log(beta + eps) - 1

    def f_inv(t: float) -> float:
        return math.pow(0.5, 1 / (t + 1))

    if step < warmup:
        a = step / float(warmup)
        return f_inv((1.0 - a) * f(beta_start) + a * f(beta_end))
    return beta_end


class AdEMAMix(Optimizer):
    r"""Implements the AdEMAMix algorithm.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999, 0.9999))
            corresponding to beta_1, beta_2, beta_3 in AdEMAMix
        alpha (float): AdEMAMix alpha coeficient mixing the slow and fast EMAs (default: 2)
        beta3_warmup (int, optional): number of warmup steps used to increase beta3 (default: None)
        alpha_warmup: (int, optional): number of warmup steps used to increase alpha (default: None)
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay as in AdamW (default: 0)
    """

    def __init__(
        self,
        params: Iterable[torch.tensor] | Iterable[dict[str, Any]],
        lr: float = 1e-3,
        betas: tuple[float, float, float] = (0.9, 0.999, 0.9999),
        alpha: float = 2.0,
        beta3_warmup: int | None = None,
        alpha_warmup: int | None = None,
        eps: float = 1e-8,
        weight_decay: float = 0.0,
    ) -> None:
        if not lr >= 0.0:
            msg = f"Invalid learning rate: {lr}"
            raise ValueError(msg)
        if not eps >= 0.0:
            msg = f"Invalid epsilon value: {eps}"
            raise ValueError(msg)
        if not 0.0 <= betas[0] < 1.0:
            msg = f"Invalid beta parameter at index 0: {betas[0]}"
            raise ValueError(msg)
        if not 0.0 <= betas[1] < 1.0:
            msg = f"Invalid beta parameter at index 1: {betas[1]}"
            raise ValueError(msg)
        if not 0.0 <= betas[2] < 1.0:
            msg = f"Invalid beta parameter at index 2: {betas[2]}"
            raise ValueError(msg)
        if not weight_decay >= 0.0:
            msg = f"Invalid weight_decay value: {weight_decay}"
            raise ValueError(msg)
        if not alpha >= 0.0:
            msg = f"Invalid alpha value: {alpha}"
            raise ValueError(msg)
        defaults = {
            "lr": lr,
            "betas": betas,
            "eps": eps,
            "alpha": alpha,
            "beta3_warmup": beta3_warmup,
            "alpha_warmup": alpha_warmup,
            "weight_decay": weight_decay,
        }
        super().__init__(params, defaults)

    @torch.no_grad()
    def step(self, closure: Callable[[], float] | None = None) -> float | None:  # noqa: C901
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:

            lr = group["lr"]
            lmbda = group["weight_decay"]
            eps = group["eps"]
            beta1, beta2, beta3_final = group["betas"]
            beta3_warmup = group["beta3_warmup"]
            alpha_final = group["alpha"]
            alpha_warmup = group["alpha_warmup"]

            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad
                if grad.is_sparse:
                    msg = "AdEMAMix does not support sparse gradients."
                    raise RuntimeError(msg)

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    if beta1 != 0.0:  # save memory in case beta1 is 0.0
                        state["exp_avg_fast"] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    else:
                        state["exp_avg_fast"] = None
                    state["exp_avg_slow"] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p, memory_format=torch.preserve_format)

                exp_avg_fast, exp_avg_slow, exp_avg_sq = (
                    state["exp_avg_fast"],
                    state["exp_avg_slow"],
                    state["exp_avg_sq"],
                )

                state["step"] += 1
                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]

                # Compute the effective alpha and beta3 in case warmup is used
                if alpha_warmup is not None:
                    alpha = linear_warmup_scheduler(
                        state["step"],
                        alpha_end=alpha_final,
                        alpha_start=0,
                        warmup=alpha_warmup,
                    )
                else:
                    alpha = alpha_final

                if beta3_warmup is not None:
                    beta3 = linear_hl_warmup_scheduler(
                        state["step"],
                        beta_end=beta3_final,
                        beta_start=beta1,
                        warmup=beta3_warmup,
                    )
                else:
                    beta3 = beta3_final

                # Decay the first and second moment running average coefficient
                if beta1 != 0.0:
                    exp_avg_fast.mul_(beta1).add_(grad, alpha=1 - beta1)
                else:
                    exp_avg_fast = grad
                exp_avg_slow.mul_(beta3).add_(grad, alpha=1 - beta3)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

                denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(eps)

                update = (exp_avg_fast.div(bias_correction1) + alpha * exp_avg_slow) / denom

                # decay
                update.add_(p, alpha=lmbda)

                p.add_(-lr * update)

        return loss



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from collections.abc import Callable
from collections.abc import Iterable
from datetime import timedelta
from typing import Any

from hydra.errors import InstantiationException
from hydra.utils import instantiate
from omegaconf import DictConfig
from pydantic import BaseModel
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.callbacks import TQDMProgressBar

from anemoi.training.diagnostics.callbacks.checkpoint import AnemoiCheckpoint
from anemoi.training.diagnostics.callbacks.optimiser import LearningRateMonitor
from anemoi.training.diagnostics.callbacks.optimiser import StochasticWeightAveraging
from anemoi.training.diagnostics.callbacks.provenance import ParentUUIDCallback
from anemoi.training.diagnostics.callbacks.sanity import CheckVariableOrder
from anemoi.training.schemas.base_schema import BaseSchema
from anemoi.training.utils.checkpoint import RegisterMigrations

LOGGER = logging.getLogger(__name__)


def nestedget(config: DictConfig, key: str, default: Any) -> Any:
    """Get a nested key from a DictConfig object.

    E.g.
    >>> nestedget(config, "diagnostics.log.wandb.enabled", False)
    """
    keys = key.split(".")
    for k in keys:
        config = getattr(config, k, default)
        if not isinstance(config, BaseModel | dict | DictConfig):
            break
    return config


# Callbacks to add according to flags in the config
# Can be function to check status from config
CONFIG_ENABLED_CALLBACKS: list[tuple[list[str] | str | Callable[[DictConfig], bool], type[Callback]]] = [
    ("training.swa.enabled", StochasticWeightAveraging),
    (
        lambda config: nestedget(config, "diagnostics.log.wandb.enabled", False)
        or nestedget(config, "diagnostics.log.mlflow.enabled", False),
        LearningRateMonitor,
    ),
]


def _get_checkpoint_callback(config: BaseSchema) -> list[AnemoiCheckpoint]:
    """Get checkpointing callbacks."""
    if not config.diagnostics.enable_checkpointing:
        return []

    checkpoint_settings = {
        "dirpath": config.system.output.checkpoints.root,
        "verbose": False,
        # save weights, optimizer states, LR-schedule states, hyperparameters etc.
        # https://pytorch-lightning.readthedocs.io/en/stable/common/checkpointing_basic.html#contents-of-a-checkpoint
        "save_weights_only": False,
        "auto_insert_metric_name": False,
        # save after every validation epoch, if we've improved
        "save_on_train_epoch_end": False,
        "enable_version_counter": False,
    }

    ckpt_frequency_save_dict = {}

    for key, frequency_dict in config.diagnostics.checkpoint.items():
        frequency = frequency_dict.save_frequency
        n_saved = frequency_dict.num_models_saved
        if key == "every_n_minutes" and frequency_dict.save_frequency is not None:
            target = "train_time_interval"
            frequency = timedelta(minutes=frequency_dict.save_frequency)
        else:
            target = key
        ckpt_frequency_save_dict[target] = (
            config.system.output.checkpoints[key],
            frequency,
            n_saved,
        )

    checkpoint_callbacks = []
    for save_key, (
        name,
        save_frequency,
        save_n_models,
    ) in ckpt_frequency_save_dict.items():
        if save_frequency is not None:
            LOGGER.debug("Checkpoint callback at %s = %s ...", save_key, save_frequency)
            checkpoint_callbacks.append(
                # save_top_k: the save_top_k flag can either save the best or the last k checkpoints
                # depending on the monitor flag on ModelCheckpoint.
                # See https://lightning.ai/docs/pytorch/stable/common/checkpointing_intermediate.html for reference
                AnemoiCheckpoint(
                    config=config,
                    filename=name,
                    save_last=True,
                    **{save_key: save_frequency},
                    # if save_top_k == k, last k models saved; if save_top_k == -1, all models are saved
                    save_top_k=save_n_models,
                    monitor="step",
                    mode="max",
                    **checkpoint_settings,
                ),
            )
        LOGGER.debug("Not setting up a checkpoint callback with %s", save_key)

    return checkpoint_callbacks


def _get_config_enabled_callbacks(config: DictConfig) -> list[Callback]:
    """Get callbacks that are enabled in the config as according to CONFIG_ENABLED_CALLBACKS."""
    callbacks = []

    def check_key(config: dict, key: str | Iterable[str] | Callable[[DictConfig], bool]) -> bool:
        """Check key in config."""
        if isinstance(key, Callable):
            return key(config)
        if isinstance(key, str):
            return nestedget(config, key, False)
        if isinstance(key, Iterable):
            return all(nestedget(config, k, False) for k in key)
        return nestedget(config, key, False)

    for enable_key, callback_list in CONFIG_ENABLED_CALLBACKS:
        if check_key(config, enable_key):
            callbacks.append(callback_list(config))

    return callbacks


def _get_progress_bar_callback(config: DictConfig) -> list[Callback]:
    """Get progress bar callback.

    Instantiated from `config.diagnostics.progress_bar`. If not set, defaults to TQDMProgressBar.

    Example config:
        progress_bar:
          _target_: pytorch_lightning.callbacks.TQDMProgressBar
          refresh_rate: 1
          process_position: 0

    Parameters
    ----------
    config : DictConfig
        Job configuration

    Returns
    -------
    list[Callback]
        List containing the progress bar callback, or empty list if disabled.
    """
    if not config.diagnostics.enable_progress_bar:
        LOGGER.info("Progress bar disabled.")
        return []

    progress_bar_cfg = nestedget(config, "diagnostics.progress_bar", None)
    if progress_bar_cfg is not None:
        try:
            progress_bar = instantiate(progress_bar_cfg)
            LOGGER.info("Using progress bar: %s", type(progress_bar))
        except InstantiationException:
            LOGGER.warning("Failed to instantiate progress bar callback from config: %s", progress_bar_cfg)
            progress_bar = TQDMProgressBar(refresh_rate=1, process_position=0)
    else:
        LOGGER.info("Using default progress bar: TQDMProgressBar.")
        progress_bar = TQDMProgressBar(refresh_rate=1, process_position=0)

    return [progress_bar]


def get_callbacks(config: DictConfig) -> list[Callback]:
    """Setup callbacks for PyTorch Lightning trainer.

    Set `config.diagnostics.callbacks` to a list of callback configurations
    in hydra form.

    E.g.:
    ```
    callbacks:
        - _target_: anemoi.training.diagnostics.callbacks.RolloutEval
          rollout: 1
          frequency: 12
    ```

    Set `config.diagnostics.plot.callbacks` to a list of plot callback configurations
    will only be added if `config.diagnostics.plot.enabled` is set to True.

    A callback must take a `DictConfig` in its `__init__` method as the first argument,
    which will be the complete configuration object.

    Some callbacks are added by default, depending on the configuration.
    See CONFIG_ENABLED_CALLBACKS for more information.

    Parameters
    ----------
    config : DictConfig
        Job configuration

    Returns
    -------
    list[Callback]
        A list of PyTorch Lightning callbacks

    """
    trainer_callbacks: list[Callback] = []

    # Get Checkpoint callback
    trainer_callbacks.extend(_get_checkpoint_callback(config))

    # Base callbacks
    trainer_callbacks.extend(instantiate(callback, config) for callback in config.diagnostics.callbacks)

    # Plotting callbacks
    trainer_callbacks.extend(instantiate(callback, config) for callback in config.diagnostics.plot.callbacks)

    # Extend with config enabled callbacks
    trainer_callbacks.extend(_get_config_enabled_callbacks(config))

    # Progress bar callback
    trainer_callbacks.extend(_get_progress_bar_callback(config))

    # Parent UUID callback
    # Check variable order callback
    # Register Migrations callback
    trainer_callbacks.extend(
        (
            ParentUUIDCallback(config),
            CheckVariableOrder(),
            RegisterMigrations(),
        ),
    )

    return trainer_callbacks


__all__ = ["get_callbacks"]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import tempfile
from pathlib import Path
from unittest import mock

import pytest
from omegaconf import OmegaConf

from anemoi.training.commands.config import ConfigGenerator


@pytest.fixture
def config_generator() -> ConfigGenerator:
    return ConfigGenerator()


def test_dump_config(config_generator: ConfigGenerator) -> None:
    with tempfile.TemporaryDirectory() as tmpdirname:
        config_path = Path(tmpdirname) / "config"
        config_path.mkdir(parents=True, exist_ok=True)
        (config_path / "test.yaml").write_text("test: value")

        output_path = Path(tmpdirname) / "output.yaml"
        with mock.patch("anemoi.training.commands.config.ConfigGenerator.copy_files") as mock_copy_files, mock.patch(
            "anemoi.training.commands.config.initialize",
        ), mock.patch("anemoi.training.commands.config.compose", return_value=OmegaConf.create({"test": "value"})):
            config_generator.dump_config(config_path, "test", output_path)

            mock_copy_files.assert_called_once_with(config_path, mock.ANY)
            assert output_path.exists()
            assert OmegaConf.load(output_path) == {"test": "value"}


def test_validate_config_uses_package_path(config_generator: ConfigGenerator) -> None:
    """Test that validate_config works correctly with package configs.

    This test verifies the fix for issue #570: the AnemoiSearchPathPlugin
    adds 'pkg://anemoi.training/config' to the search path, enabling
    discovery of package configs like 'training/default'.

    Note: The package path is added by the plugin, not by the initialize() call,
    so we just verify the method executes successfully.
    """
    with mock.patch("anemoi.training.commands.config.initialize"), mock.patch(
        "anemoi.training.commands.config.compose",
        return_value=OmegaConf.create({"test": "value"}),
    ), mock.patch("anemoi.training.commands.config.BaseSchema"):
        # Call validate_config - it should succeed
        # The AnemoiSearchPathPlugin will add the package path automatically
        config_generator.validate_config("test-config", mask_env_vars=False)

        # If we get here without exception, the validation worked


def test_validate_config_with_mask_env_vars(config_generator: ConfigGenerator) -> None:
    """Test that validate_config works with mask_env_vars option."""
    with mock.patch("anemoi.training.commands.config.initialize"), mock.patch(
        "anemoi.training.commands.config.compose",
        return_value=OmegaConf.create({"test": "value"}),
    ), mock.patch("anemoi.training.commands.config.BaseSchema"), mock.patch.object(
        config_generator,
        "_mask_slurm_env_variables",
        return_value=OmegaConf.create({"test": "masked"}),
    ) as mock_mask:
        # Call validate_config with mask_env_vars=True
        config_generator.validate_config("test-config", mask_env_vars=True)

        # Verify that _mask_slurm_env_variables was called
        mock_mask.assert_called_once()



# (C) Copyright 2024- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
import sys
from pathlib import Path
from typing import Any
from typing import Union

from omegaconf import DictConfig
from omegaconf import OmegaConf
from pydantic import BaseModel as PydanticBaseModel
from pydantic import model_validator
from pydantic._internal import _model_construction
from pydantic_core import PydanticCustomError
from pydantic_core import ValidationError
from typing_extensions import Self

from anemoi.graphs.schemas.base_graph import BaseGraphSchema
from anemoi.models.schemas.decoder import GraphTransformerDecoderSchema
from anemoi.models.schemas.models import ModelSchema
from anemoi.utils.schemas import BaseModel
from anemoi.utils.schemas.errors import CUSTOM_MESSAGES
from anemoi.utils.schemas.errors import convert_errors

# to make these available at runtime for pydantic, bug should be resolved in
# future versions (see https://github.com/astral-sh/ruff/issues/7866)
from .data import DataSchema
from .dataloader import DataLoaderSchema
from .diagnostics import DiagnosticsSchema
from .system import SystemSchema
from .training import TrainingSchema

_object_setattr = _model_construction.object_setattr

LOGGER = logging.getLogger(__name__)


def expand_paths(config_system: Union[SystemSchema, DictConfig]) -> Union[SystemSchema, DictConfig]:
    output_config = config_system.output
    root_output_path = Path(output_config.root) if output_config.root else Path()
    # OutputSchema
    if output_config.plots:
        config_system.output.plots = root_output_path / output_config.plots
    if output_config.profiler:
        config_system.output.profiler = root_output_path / output_config.profiler

    # LogsSchema
    config_system.output.logs.root = (
        root_output_path / output_config.logs.root if output_config.logs.root else root_output_path
    )
    base = config_system.output.logs.root

    # LogsSchema
    output_config.logs.wandb = base / "wandb" if output_config.logs.wandb is None else base / output_config.logs.wandb
    output_config.logs.mlflow = (
        base / "mlflow" if output_config.logs.mlflow is None else base / output_config.logs.mlflow
    )
    output_config.logs.tensorboard = (
        base / "tensorboard" if output_config.logs.tensorboard is None else base / output_config.logs.tensorboard
    )

    # CheckPointSchema
    output_config.checkpoints.root = (
        root_output_path / output_config.checkpoints.root if output_config.checkpoints.root else root_output_path
    )

    return config_system


class SchemaCommonMixin:
    """Shared logic for schema objects."""

    def model_dump(self, by_alias: bool = False) -> dict:
        dumped_model = super().model_dump(by_alias=by_alias)
        return DictConfig(dumped_model)

    def model_post_init(self, _: Any) -> None:
        expand_paths(self.system)
        if self.diagnostics.log.mlflow.enabled and (
            self.system.output.logs.mlflow != self.diagnostics.log.mlflow.save_dir
        ):
            LOGGER.info("adjusting save_dir path to match output mlflow logs")
            self.diagnostics.log.mlflow.save_dir = str(self.system.output.logs.mlflow)


class BaseSchema(SchemaCommonMixin, BaseModel):
    """Top-level schema for the training configuration."""

    data: DataSchema
    """Data configuration."""
    dataloader: DataLoaderSchema
    """Dataloader configuration."""
    diagnostics: DiagnosticsSchema
    """Diagnostics configuration such as logging, plots and metrics."""
    system: SystemSchema
    """System configuration, including filesystem and hardware specification."""
    graph: BaseGraphSchema
    """Graph configuration."""
    model: ModelSchema
    """Model configuration."""
    training: TrainingSchema
    """Training configuration."""
    config_validation: bool = True
    """Flag to disable validation of the configuration"""

    @model_validator(mode="after")
    def set_read_group_size_if_not_provided(self) -> Self:
        if not self.dataloader.read_group_size:
            self.dataloader.read_group_size = self.system.hardware.num_gpus_per_model
        return self

    @model_validator(mode="after")
    def check_bounding_not_used_with_data_extractor_zero(self) -> Self:
        """Check that bounding is not used with zero data extractor."""
        if (
            isinstance(self.model.decoder, GraphTransformerDecoderSchema)
            and self.model.decoder.initialise_data_extractor_zero
            and self.model.bounding
        ):
            error = "bounding_conflict_with_data_extractor_zero"
            msg = (
                "Boundings cannot be used with zero initialized weights in decoder. "
                "Set initalise_data_extractor_zero to False."
            )
            raise PydanticCustomError(
                error,
                msg,
            )
        return self


class UnvalidatedBaseSchema(SchemaCommonMixin, PydanticBaseModel):
    data: Any
    """Data configuration."""
    dataloader: Any
    """Dataloader configuration."""
    diagnostics: Any
    """Diagnostics configuration such as logging, plots and metrics."""
    system: Any
    """Hardware configuration."""
    graph: Any
    """Graph configuration."""
    model: Any
    """Model configuration."""
    training: Any
    """Training configuration."""
    config_validation: bool = False
    """Flag to disable validation of the configuration"""


def convert_to_omegaconf(config: BaseSchema) -> dict:
    config = config.model_dump(by_alias=True)
    return OmegaConf.create(config)


def validate_schema(config: DictConfig) -> BaseSchema:
    try:
        config = BaseSchema(**config)
    except ValidationError as e:
        errors = convert_errors(e, CUSTOM_MESSAGES)
        LOGGER.error(errors)  # noqa: TRY400
        sys.exit(0)
    else:
        return config



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import numpy as np


def get_usable_indices(
    missing_indices: set[int],
    series_length: int,
    relative_indices: np.ndarray,
    trajectory_ids: np.ndarray | None = None,
) -> np.ndarray:
    """Get the usable indices of a series with missing indices.

    Parameters
    ----------
    missing_indices : set[int]
        Dataset to be used.
    series_length : int
        Length of the series.
    relative_indices: array[np.int64]
        Array of relative indices requested at each index i.
    trajectory_ids: array[np.int64]
        Array of integers of length series length that indicates which forecast trajectory a time index belongs to.
        When training on analysis: None

    Returns
    -------
    usable_indices : np.array
        Array of usable indices.
    """
    usable_indices = np.arange(series_length - max(relative_indices))

    # Avoid crossing model runs by selecting only relative indices with the same model run id
    if trajectory_ids is not None:
        rel_run = usable_indices[None] + relative_indices[:, None]
        include = (trajectory_ids[rel_run] == trajectory_ids[rel_run[0]]).all(axis=0)
        usable_indices = usable_indices[include]

    # Missing indices
    for i in missing_indices:
        rel_missing = i - relative_indices  # indices which have their relative indices match the missing.
        usable_indices = usable_indices[np.all(usable_indices != rel_missing[:, np.newaxis], axis=0)]

    return usable_indices



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from .combined import CombinedLoss
from .huber import HuberLoss
from .kcrps import AlmostFairKernelCRPS
from .kcrps import KernelCRPS
from .logcosh import LogCoshLoss
from .loss import get_loss_function
from .mae import MAELoss
from .mse import MSELoss
from .multiscale import MultiscaleLossWrapper
from .rmse import RMSELoss
from .weighted_mse import WeightedMSELoss

__all__ = [
    "AlmostFairKernelCRPS",
    "CombinedLoss",
    "HuberLoss",
    "KernelCRPS",
    "LogCoshLoss",
    "MAELoss",
    "MSELoss",
    "MultiscaleLossWrapper",
    "RMSELoss",
    "WeightedMSELoss",
    "get_loss_function",
]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from abc import ABC
from abc import abstractmethod
from functools import cached_property
from pathlib import Path
from typing import Any

import hydra
import numpy as np
import pytorch_lightning as pl
import torch
from hydra.utils import get_class
from hydra.utils import instantiate
from omegaconf import DictConfig
from omegaconf import OmegaConf
from packaging import version
from pytorch_lightning.utilities.rank_zero import rank_zero_only
from torch_geometric.data import HeteroData

from anemoi.models.utils.compile import mark_for_compilation
from anemoi.training.data.datamodule import AnemoiDatasetsDataModule
from anemoi.training.diagnostics.callbacks import get_callbacks
from anemoi.training.diagnostics.logger import get_mlflow_logger
from anemoi.training.diagnostics.logger import get_tensorboard_logger
from anemoi.training.diagnostics.logger import get_wandb_logger
from anemoi.training.schemas.base_schema import BaseSchema
from anemoi.training.schemas.base_schema import UnvalidatedBaseSchema
from anemoi.training.schemas.base_schema import convert_to_omegaconf
from anemoi.training.utils.checkpoint import freeze_submodule_by_name
from anemoi.training.utils.checkpoint import transfer_learning_loading
from anemoi.training.utils.jsonify import map_config_to_primitives
from anemoi.training.utils.seeding import get_base_seed
from anemoi.utils.provenance import gather_provenance_info

LOGGER = logging.getLogger(__name__)

PL_VERSION = version.parse(pl.__version__)


class AnemoiTrainer(ABC):
    """Utility class for training the model."""

    def __init__(self, config: DictConfig) -> None:
        """Initialize the Anemoi trainer.

        Parameters
        ----------
        config : DictConfig
            Config object from Hydra.

        """
        # Allow for lower internal precision of float32 matrix multiplications.
        # This can increase performance (and TensorCore usage, where available).
        torch.set_float32_matmul_precision("high")
        # Resolve the config to avoid shenanigans with lazy loading

        if config.config_validation:
            OmegaConf.resolve(config)
            self.config = BaseSchema(**config)

            LOGGER.info("Config validated.")
        else:
            config = OmegaConf.to_object(config)
            self.config = UnvalidatedBaseSchema(**DictConfig(config))

            LOGGER.info("Skipping config validation.")

        self.start_from_checkpoint = (
            bool(self.config.training.run_id)
            or bool(self.config.training.fork_run_id)
            or bool(self.config.system.input.warm_start)
        )
        LOGGER.info("Starting from checkpoint: %s", self.start_from_checkpoint)

        self.load_weights_only = self.config.training.load_weights_only
        self.parent_uuid = None

        self.config.training.run_id = self.run_id
        LOGGER.info("Run id: %s", self.config.training.run_id)

        # Get the server2server lineage
        self._get_server2server_lineage()

        # Update paths to contain the run ID
        self._update_paths()

        # Update dry_run attribute, check if checkpoint exists
        self._check_dry_run()

        # Check for dry run, i.e. run id without data
        self._log_information()

    @cached_property
    def datamodule(self) -> Any:
        """DataModule instance and DataSets."""
        datamodule = AnemoiDatasetsDataModule(convert_to_omegaconf(self.config), self.graph_data)
        self.config.data.num_features = len(datamodule.ds_train.data.variables)
        LOGGER.info("Number of data variables: %s", str(len(datamodule.ds_train.data.variables)))
        LOGGER.info("Variables: %s", str(datamodule.ds_train.data.variables))
        return datamodule

    @cached_property
    def data_indices(self) -> dict:
        """Returns a dictionary of data indices.

        This is used to slice the data.
        """
        return self.datamodule.data_indices

    @cached_property
    def initial_seed(self) -> int:
        """Initial seed for the RNG.

        This sets the same initial seed for all ranks. Ranks are re-seeded in the
        strategy to account for model communication groups.
        """
        initial_seed = get_base_seed()
        rnd_seed = pl.seed_everything(initial_seed, workers=True)
        np_rng = np.random.default_rng(rnd_seed)
        (torch.rand(1), np_rng.random())
        LOGGER.info(
            "Initial seed: Rank %d, initial seed %d, running with random seed: %d",
            self.strategy.global_rank,
            initial_seed,
            rnd_seed,
        )
        return initial_seed

    @cached_property
    def graph_data(self) -> HeteroData:
        """Graph data.

        Creates the graph in all workers.
        """
        if (graph_filename := self.config.system.input.graph) is not None:
            graph_filename = Path(graph_filename)
            if graph_filename.exists() and not self.config.graph.overwrite:
                from anemoi.graphs.utils import get_distributed_device

                LOGGER.info("Loading graph data from %s", graph_filename)
                return torch.load(graph_filename, map_location=get_distributed_device(), weights_only=False)

        else:
            graph_filename = None

        from anemoi.graphs.create import GraphCreator

        graph_config = convert_to_omegaconf(self.config).graph
        return GraphCreator(config=graph_config).create(
            save_path=graph_filename,
            overwrite=self.config.graph.overwrite,
        )

    @cached_property
    @abstractmethod
    def profiler(self) -> None:
        """Abstract method to be used for AnemoiProfiler."""
        return None

    @cached_property
    def model(self) -> pl.LightningModule:
        """Provide the model instance."""
        assert (
            not (
                "GLU" in self.config.model.processor.layer_kernels["Activation"]["_target_"]
                and ".Transformer" in self.config.model.processor.target_
            )
            and not (
                "GLU" in self.config.model.encoder.layer_kernels["Activation"]["_target_"]
                and ".Transformer" in self.config.model.encoder.target_
            )
            and not (
                "GLU" in self.config.model.decoder.layer_kernels["Activation"]["_target_"]
                and ".Transformer" in self.config.model.decoder.target_
            )
        ), "GLU activation function is not supported in Transformer models, due to fixed dimensions. "
        "Please use a different activation function."

        kwargs = {
            "config": self.config,
            "data_indices": self.data_indices,
            "graph_data": self.graph_data,
            "metadata": self.metadata,
            "statistics": self.datamodule.statistics,
            "statistics_tendencies": self.datamodule.statistics_tendencies,
            "supporting_arrays": self.supporting_arrays,
        }

        model_task = get_class(self.config.training.model_task)
        model = model_task(**kwargs)  # GraphForecaster -> pl.LightningModule

        # Load the model weights
        if self.load_weights_only:
            # Sanify the checkpoint for transfer learning
            if self.config.training.transfer_learning:
                LOGGER.info("Loading weights with Transfer Learning from %s", self.last_checkpoint)
                model = transfer_learning_loading(model, self.last_checkpoint)
            else:
                LOGGER.info("Restoring only model weights from %s", self.last_checkpoint)
                # pop data_indices so that the data indices on the checkpoint do not get overwritten
                # by the data indices from the new config
                kwargs.pop("data_indices")
                model = model_task.load_from_checkpoint(
                    self.last_checkpoint,
                    **kwargs,
                    strict=False,
                    weights_only=False,  # required for Pytorch Lightning 2.6
                )

            model.data_indices = self.data_indices
            # check data indices in original checkpoint and current data indices are the same
            self.data_indices.compare_variables(model._ckpt_model_name_to_index, self.data_indices.name_to_index)

        if hasattr(self.config.training, "submodules_to_freeze"):
            # Freeze the chosen model weights
            LOGGER.info("The following submodules will NOT be trained: %s", self.config.training.submodules_to_freeze)
            for submodule_name in self.config.training.submodules_to_freeze:
                freeze_submodule_by_name(model, submodule_name)
                LOGGER.info("%s frozen successfully.", submodule_name.upper())

        return model

    @rank_zero_only
    def _get_mlflow_run_id(self) -> str:
        run_id = self.mlflow_logger.run_id
        # for resumed runs or offline runs logging this can be useful
        LOGGER.info("Mlflow Run id: %s", run_id)
        return run_id

    @cached_property
    def run_id(self) -> str:
        """Unique identifier for the current run."""
        # When a run ID is provided
        if self.config.training.run_id and not self.config.training.fork_run_id:
            # Return the provided run ID - reuse run_id if resuming run
            return self.config.training.run_id

        # When a run ID has been created externally and we want to fork a run
        if self.config.training.run_id and self.config.training.fork_run_id:
            return self.config.training.run_id

        # When we rely on mlflow to create a new run ID
        if self.config.diagnostics.log.mlflow.enabled:
            # if using mlflow with a new run get the run_id from mlflow
            return self._get_mlflow_run_id()

        # When no run ID is provided a random one is generated
        import uuid

        return str(uuid.uuid4())

    @cached_property
    def wandb_logger(self) -> pl.loggers.WandbLogger:
        """WandB logger."""
        return get_wandb_logger(self.config, self.model)

    @cached_property
    def mlflow_logger(self) -> pl.loggers.MLFlowLogger:
        """Mlflow logger."""
        return get_mlflow_logger(self.config)

    @cached_property
    def tensorboard_logger(self) -> pl.loggers.TensorBoardLogger:
        """TensorBoard logger."""
        return get_tensorboard_logger(self.config)

    def _get_warm_start_checkpoint(self) -> Path | None:
        """Returns the warm start checkpoint path if specified."""
        raw_path = self.config.system.input.warm_start
        if not raw_path:
            return None

        warm_start_path = Path(raw_path)

        if not warm_start_path.is_file():
            msg = f"Warm start checkpoint not found: {warm_start_path}"
            raise FileNotFoundError(msg)
        return warm_start_path

    def _get_checkpoint_directory(self, fork_id: str) -> Path:
        """Returns the directory where checkpoints are stored."""
        return Path(self.config.system.output.checkpoints.root.parent, fork_id or self.lineage_run) / "last.ckpt"

    @cached_property
    def last_checkpoint(self) -> Path | None:
        """Path to the last checkpoint."""
        if not self.start_from_checkpoint:
            return None

        fork_id = self.fork_run_server2server or self.config.training.fork_run_id
        checkpoint = self._get_warm_start_checkpoint() or self._get_checkpoint_directory(fork_id)
        # Check if the last checkpoint exists
        if checkpoint.exists():
            LOGGER.info("Resuming training from last checkpoint: %s", checkpoint)
            return checkpoint

        if rank_zero_only.rank == 0:
            msg = "Could not find last checkpoint: %s", checkpoint
            raise RuntimeError(msg)

        return None

    @cached_property
    def callbacks(self) -> list[pl.callbacks.Callback]:
        return get_callbacks(self.config.model_dump(by_alias=True))

    @cached_property
    def metadata(self) -> dict:
        """Metadata and provenance information."""
        return map_config_to_primitives(
            {
                "version": "1.0",
                "config": convert_to_omegaconf(self.config),
                "seed": self.initial_seed,
                "run_id": self.run_id,
                "dataset": self.datamodule.metadata,
                "data_indices": self.datamodule.data_indices,
                "provenance_training": gather_provenance_info(),
                "timestamp": datetime.datetime.now(tz=datetime.timezone.utc),
            },
        )

    @cached_property
    def supporting_arrays(self) -> dict:
        return self.datamodule.supporting_arrays

    @cached_property
    def loggers(self) -> list:
        loggers = []
        if self.config.diagnostics.log.wandb.enabled:
            LOGGER.info("W&B logger enabled")
            loggers.append(self.wandb_logger)
        if self.config.diagnostics.log.tensorboard.enabled:
            LOGGER.info("TensorBoard logger enabled")
            loggers.append(self.tensorboard_logger)
        if self.config.diagnostics.log.mlflow.enabled:
            LOGGER.info("MLFlow logger enabled")
            loggers.append(self.mlflow_logger)
        return loggers

    @cached_property
    def accelerator(self) -> str:
        assert self.config.system.hardware.accelerator in {
            "auto",
            "cpu",
            "gpu",
            "cuda",
            "tpu",
        }, f"Invalid accelerator ({self.config.system.hardware.accelerator}) in system.hardware config."

        if self.config.system.hardware.accelerator == "cpu":
            LOGGER.info("WARNING: Accelerator set to CPU, this should only be used for debugging.")
        return self.config.system.hardware.accelerator

    def _log_information(self) -> None:
        # Log number of variables (features)
        num_fc_features = len(self.datamodule.ds_train.data.variables) - len(self.config.data.forcing)
        LOGGER.info("Total number of prognostic variables: %d", num_fc_features)
        LOGGER.info("Total number of auxiliary variables: %d", len(self.config.data.forcing))

        # Log learning rate multiplier when running single-node, multi-GPU and/or multi-node
        total_number_of_model_instances = (
            self.config.system.hardware.num_nodes
            * self.config.system.hardware.num_gpus_per_node
            / self.config.system.hardware.num_gpus_per_model
        )

        LOGGER.info(
            "Total GPU count / model group size: %d - NB: the learning rate will be scaled by this factor!",
            total_number_of_model_instances,
        )
        LOGGER.info(
            "Effective learning rate: %.3e",
            int(total_number_of_model_instances) * self.config.training.lr.rate,
        )

        if self.config.training.max_epochs is not None and self.config.training.max_steps not in (None, -1):
            LOGGER.info(
                "Training limits: max_epochs=%d, max_steps=%d. "
                "Training will stop when either limit is reached first. "
                "Learning rate scheduler will run for %d steps.",
                self.config.training.max_epochs,
                self.config.training.max_steps,
                self.config.training.lr.iterations,
            )

    def _get_server2server_lineage(self) -> None:
        """Get the server2server lineage."""
        self.parent_run_server2server = None
        self.fork_run_server2server = None
        if self.config.diagnostics.log.mlflow.enabled:
            self.parent_run_server2server = self.mlflow_logger._parent_run_server2server
            LOGGER.info("Parent run server2server: %s", self.parent_run_server2server)
            self.fork_run_server2server = self.mlflow_logger._fork_run_server2server
            LOGGER.info("Fork run server2server: %s", self.fork_run_server2server)

    def _update_paths(self) -> None:
        """Update the paths in the configuration."""
        self.lineage_run = None
        if self.run_id:  # when using mlflow only rank0 will have a run_id except when resuming runs
            # Multi-gpu new runs or forked runs - only rank 0
            # Multi-gpu resumed runs - all ranks
            self.lineage_run = self.parent_run_server2server or self.run_id
            self.config.system.output.checkpoints.root = Path(
                self.config.system.output.checkpoints.root,
                self.lineage_run,
            )
            self.config.system.output.plots = Path(self.config.system.output.plots, self.lineage_run)
        elif self.config.training.fork_run_id:
            # WHEN USING MANY NODES/GPUS
            self.lineage_run = self.parent_run_server2server or self.config.training.fork_run_id
            # Only rank non zero in the forked run will go here
            self.config.system.output.checkpoints.root = Path(
                self.config.system.output.checkpoints.root,
                self.lineage_run,
            )

        LOGGER.info("Checkpoints path: %s", self.config.system.output.checkpoints)
        LOGGER.info("Plots path: %s", self.config.system.output.plots)

    @rank_zero_only
    def _check_dry_run(self) -> None:
        """Check if the run ID is dry, e.g. without a checkpoint.

        If the run ID is dry, the training will not be started.
        This is used to check the run can be restarted from the checkpoint.
        """
        self.dry_run = False
        if self.config.diagnostics.log.mlflow.enabled:
            # Check if the run ID is dry - e.g. without a checkpoint
            self.dry_run = (
                self.mlflow_logger._parent_dry_run and not Path(self.config.system.output.checkpoints.root).is_dir()
            )
            self.start_from_checkpoint = (
                False if (self.dry_run and not bool(self.config.training.fork_run_id)) else self.start_from_checkpoint
            )
            LOGGER.info("Dry run: %s", self.dry_run)

    def prepare_compilation(self) -> None:

        if hasattr(self.config.model, "compile"):
            self.model = mark_for_compilation(self.model, self.config.model_dump(by_alias=True).model.compile)
        if hasattr(self.config.training, "recompile_limit"):
            torch._dynamo.config.cache_size_limit = int(self.config.training.recompile_limit)
            torch._dynamo.config.accumulated_cache_size_limit = max(8 * int(self.config.training.recompile_limit), 256)
            LOGGER.info("Recompile limit set to %d", torch._dynamo.config.cache_size_limit)

    @cached_property
    def strategy(self) -> Any:
        return instantiate(
            convert_to_omegaconf(self.config).training.strategy,
            static_graph=not self.config.training.accum_grad_batches > 1,
        )

    @cached_property
    def fit_parameters(self) -> Any:
        """Options to be passed to trainer.fit().

        This builds up different arguments based on the version of pytorch lightning.
        From 2.6 onwards pytorch-lightning has now exposed the weights_only flag to be
        consistent with Pytorch's behaviour.
        Refer to https://docs.pytorch.org/docs/stable/generated/torch.load.html for more details.
        `weights_only` does not refer to loading the optimizer. Pytorch_lightning controls this
        via the checkpoint connector. If a ckpt_path is passed then all states are loaded. If no ckpt_path
        is passed and just the `load_from_checkpoint` interface is used - then optimizer states are skipped.
        """
        params = {}

        params["model"] = self.model
        params["datamodule"] = self.datamodule
        params["ckpt_path"] = None if (self.load_weights_only) else self.last_checkpoint

        if version.parse("2.6.0") <= PL_VERSION:
            params["weights_only"] = False
        return params

    def train(self) -> None:
        """Training entry point."""
        LOGGER.debug("Setting up trainer..")

        trainer = pl.Trainer(
            accelerator=self.accelerator,
            callbacks=self.callbacks,
            deterministic=self.config.training.deterministic,
            detect_anomaly=self.config.diagnostics.debug.anomaly_detection,
            strategy=self.strategy,
            devices=self.config.system.hardware.num_gpus_per_node,
            num_nodes=self.config.system.hardware.num_nodes,
            precision=self.config.training.precision,
            max_epochs=self.config.training.max_epochs,
            max_steps=self.config.training.max_steps or -1,
            logger=self.loggers,
            profiler=self.profiler,
            log_every_n_steps=self.config.diagnostics.log.interval,
            # run a fixed no of batches per epoch (helpful when debugging)
            limit_train_batches=self.config.dataloader.limit_batches.training,
            limit_val_batches=self.config.dataloader.limit_batches.validation,
            num_sanity_val_steps=self.config.training.num_sanity_val_steps,
            accumulate_grad_batches=self.config.training.accum_grad_batches,
            gradient_clip_val=self.config.training.gradient_clip.val,
            gradient_clip_algorithm=self.config.training.gradient_clip.algorithm,
            # we have our own DDP-compliant sampler logic baked into the dataset
            use_distributed_sampler=False,
            enable_progress_bar=self.config.diagnostics.enable_progress_bar,
            check_val_every_n_epoch=getattr(self.config.diagnostics, "check_val_every_n_epoch", 1),
        )

        self.prepare_compilation()

        LOGGER.debug("Starting training..")

        trainer.fit(**self.fit_parameters)

        if self.config.diagnostics.print_memory_summary:
            LOGGER.info("memory summary: %s", torch.cuda.memory_summary(device=0))

        LOGGER.debug("---- DONE. ----")


@hydra.main(version_base=None, config_path="../config", config_name="config")
def main(config: DictConfig) -> None:
    AnemoiTrainer(config).train()


if __name__ == "__main__":
    main()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import TYPE_CHECKING

import pytorch_lightning as pl

if TYPE_CHECKING:
    from anemoi.training.data.dataset import NativeGridDataset

LOGGER = logging.getLogger(__name__)


class CheckVariableOrder(pl.callbacks.Callback):
    """Check the order of the variables in a pre-trained / fine-tuning model."""

    def __init__(self) -> None:
        super().__init__()

    def _check_variable_order(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        dataset: "NativeGridDataset",
    ) -> None:
        """Check the order of variables between model checkpoint and dataset.

        Parameters
        ----------
        trainer : pl.Trainer
            Pytorch Lightning trainer
        pl_module : pl.LightningModule
            Lightning module (already unwrapped by PyTorch Lightning)
        dataset : NativeGridDataset
            Dataset to compare against
        """
        data_name_to_index = dataset.name_to_index

        if hasattr(pl_module, "_ckpt_model_name_to_index"):
            model_name_to_index = pl_module._ckpt_model_name_to_index
        else:
            model_name_to_index = trainer.datamodule.data_indices.name_to_index

        trainer.datamodule.data_indices.compare_variables(model_name_to_index, data_name_to_index)

    def on_train_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Check the order of the variables in the model from checkpoint and the training data.

        Parameters
        ----------
        trainer : pl.Trainer
            Pytorch Lightning trainer
        pl_module : pl.LightningModule
            Lightning module
        """
        self._check_variable_order(trainer, pl_module, trainer.datamodule.ds_train)

    def on_validation_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Check the order of the variables in the model from checkpoint and the validation data.

        Parameters
        ----------
        trainer : pl.Trainer
            Pytorch Lightning trainer
        pl_module : pl.LightningModule
            Lightning module
        """
        self._check_variable_order(trainer, pl_module, trainer.datamodule.ds_valid)

    def on_test_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Check the order of the variables in the model from checkpoint and the test data.

        Parameters
        ----------
        trainer : pl.Trainer
            Pytorch Lightning trainer
        pl_module : pl.LightningModule
            Lightning module
        """
        self._check_variable_order(trainer, pl_module, trainer.datamodule.ds_test)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import os
from pathlib import Path
from typing import Union

import pytest
import torch
from hydra import compose
from hydra import initialize
from omegaconf import DictConfig
from omegaconf import ListConfig
from omegaconf import OmegaConf

from anemoi.models.migrations import Migrator
from anemoi.utils.testing import GetTestData
from anemoi.utils.testing import TemporaryDirectoryForTestData

LOGGER = logging.getLogger(__name__)


@pytest.fixture(autouse=True)
def set_working_directory() -> None:
    """Automatically set the working directory to the repo root."""
    repo_root = Path(__file__).resolve().parent
    while not (repo_root / ".git").exists() and repo_root != repo_root.parent:
        repo_root = repo_root.parent

    os.chdir(repo_root)


def _load_testing_modifications(tmp_path: Path) -> Union[DictConfig, ListConfig]:
    modifications_file = "training/tests/integration/config/testing_modifications.yaml"
    testing_modifications = OmegaConf.load(Path.cwd() / modifications_file)
    assert isinstance(testing_modifications, DictConfig)
    testing_modifications.system.output.root = str(tmp_path)
    return testing_modifications


@pytest.fixture
def testing_modifications_with_temp_dir(tmp_path: Path) -> DictConfig:
    return _load_testing_modifications(tmp_path)


@pytest.fixture
def testing_modifications_callbacks_on_with_temp_dir(tmp_path: Path) -> DictConfig:
    testing_modifications = _load_testing_modifications(tmp_path)
    del testing_modifications.diagnostics.plot.callbacks
    return testing_modifications


class GetTmpPaths:
    def __init__(self, temporary_directory_for_test_data: TemporaryDirectoryForTestData) -> None:
        self.temporary_directory_for_test_data = temporary_directory_for_test_data

    def __call__(self, config: DictConfig, list_datasets: list[str]) -> tuple[str, list[str], list[str]]:
        tmp_paths = []
        dataset_names = []
        archive_urls = []

        for dataset in list_datasets:
            url_archive = config.system.input[dataset] + ".tgz"
            name_dataset = Path(config.system.input[dataset]).name
            tmp_path_dataset = self.temporary_directory_for_test_data(url_archive, archive=True)

            tmp_paths.append(tmp_path_dataset)
            dataset_names.append(name_dataset)
            archive_urls.append(url_archive)

        if len(list_datasets) == 1:
            return tmp_paths[0], dataset_names, archive_urls

        tmp_dir = os.path.commonprefix([tmp_paths[0], tmp_paths[1]])[:-1]  # remove trailing slash
        rel_paths = [Path(Path(path).name) / name for (name, path) in zip(dataset_names, tmp_paths, strict=False)]
        return tmp_dir, rel_paths, archive_urls


@pytest.fixture
def get_tmp_paths(temporary_directory_for_test_data: TemporaryDirectoryForTestData) -> GetTmpPaths:
    return GetTmpPaths(temporary_directory_for_test_data)


@pytest.fixture(
    params=[
        ["config_validation=True", "diagnostics.log.mlflow.enabled=True", "diagnostics.log.mlflow.offline=True"],
        ["config_validation=True", "diagnostics.log.mlflow.enabled=False"],
        [
            "config_validation=False",
            "diagnostics.log.mlflow.enabled=True",
            "system.input.graph=null",
            "diagnostics.log.mlflow.offline=True",
        ],
        ["config_validation=False", "diagnostics.log.mlflow.enabled=False", "system.input.graph=null"],
    ],
)
def base_global_config(
    request: pytest.FixtureRequest,
    testing_modifications_with_temp_dir: DictConfig,
    get_tmp_paths: GetTmpPaths,
) -> tuple[DictConfig, str, str]:
    overrides = request.param
    model_architecture = overrides[0].split("=")[1]
    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_config"):
        template = compose(
            config_name="config",
            overrides=overrides,
        )  # apply architecture overrides to template since they override a default
    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_config.yaml")
    assert isinstance(use_case_modifications, DictConfig)

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.system.input.dataset = str(Path(tmp_dir, rel_paths[0]))

    # Add the imputer here as it's not part of the default config
    imputer_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/imputer_modifications.yaml")

    OmegaConf.set_struct(template.data, False)  # allow adding new keys to the template to add the imputer
    cfg = OmegaConf.merge(
        template,
        testing_modifications_with_temp_dir,
        use_case_modifications,
        imputer_modifications,
        OmegaConf.from_dotlist(overrides),
    )

    OmegaConf.resolve(cfg)
    assert isinstance(cfg, DictConfig)
    return cfg, dataset_urls[0], model_architecture


def build_architecture_config(
    overrides: list[str],
    testing_modifications: DictConfig,
    get_tmp_paths: GetTmpPaths,
) -> tuple[DictConfig, str, str]:

    model_architecture = overrides[0].split("=")[1]

    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_config"):
        template = compose(config_name="config", overrides=overrides)

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_config.yaml")

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.system.input.dataset = str(Path(tmp_dir, rel_paths[0]))

    imputer_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/imputer_modifications.yaml")

    OmegaConf.set_struct(template.data, False)
    cfg = OmegaConf.merge(
        template,
        testing_modifications,
        use_case_modifications,
        imputer_modifications,
    )
    OmegaConf.resolve(cfg)

    return cfg, dataset_urls[0], model_architecture


@pytest.fixture(
    params=[
        ["model=gnn"],
        ["model=graphtransformer"],
    ],
)
def architecture_config(
    request: pytest.FixtureRequest,
    testing_modifications_callbacks_on_with_temp_dir: DictConfig,
    get_tmp_paths: GetTmpPaths,
) -> tuple[DictConfig, str, str]:
    return build_architecture_config(
        request.param,
        testing_modifications_callbacks_on_with_temp_dir,
        get_tmp_paths,
    )


@pytest.fixture
def stretched_config(
    testing_modifications_with_temp_dir: DictConfig,
    get_tmp_paths: GetTmpPaths,
) -> tuple[DictConfig, list[str]]:
    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_stretched"):
        template = compose(config_name="stretched")

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_stretched.yaml")
    assert isinstance(use_case_modifications, DictConfig)

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset", "forcing_dataset"])
    dataset, forcing_dataset = rel_paths
    use_case_modifications.system.input.dataset = str(Path(tmp_dir, dataset))
    use_case_modifications.system.input.forcing_dataset = str(Path(tmp_dir, forcing_dataset))

    cfg = OmegaConf.merge(template, testing_modifications_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    assert isinstance(cfg, DictConfig)
    return cfg, dataset_urls


@pytest.fixture
def lam_config(
    testing_modifications_callbacks_on_with_temp_dir: DictConfig,
    get_tmp_paths: GetTmpPaths,
) -> tuple[DictConfig, list[str]]:
    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_lam"):
        template = compose(config_name="lam")

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_lam.yaml")
    assert isinstance(use_case_modifications, DictConfig)

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset", "forcing_dataset"])
    dataset, forcing_dataset = rel_paths
    use_case_modifications.system.input.dataset = str(Path(tmp_dir, dataset))
    use_case_modifications.system.input.forcing_dataset = str(Path(tmp_dir, forcing_dataset))

    cfg = OmegaConf.merge(template, testing_modifications_callbacks_on_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    assert isinstance(cfg, DictConfig)
    return cfg, dataset_urls


@pytest.fixture
def lam_config_with_graph(
    lam_config: tuple[DictConfig, list[str]],
    get_test_data: GetTestData,
) -> tuple[DictConfig, list[str]]:
    existing_graph_config = OmegaConf.load(Path.cwd() / "training/src/anemoi/training/config/graph/existing.yaml")
    cfg, urls = lam_config
    cfg.graph = existing_graph_config

    url_graph = cfg.system.input.graph
    tmp_path_graph = get_test_data(url_graph)
    cfg.system.input.graph = Path(tmp_path_graph)
    return cfg, urls


def handle_truncation_matrices(cfg: DictConfig, get_test_data: GetTestData) -> DictConfig:
    url_loss_matrices = cfg.system.input.loss_matrices_path
    tmp_path_loss_matrices = None
    for file in cfg.training.training_loss.loss_matrices:
        if file is not None:
            tmp_path_loss_matrices = get_test_data(url_loss_matrices + file)
    if tmp_path_loss_matrices is not None:
        cfg.system.input.loss_matrices_path = Path(tmp_path_loss_matrices).parent
        cfg.training.training_loss.loss_matrices_path = str(Path(tmp_path_loss_matrices).parent)
        cfg.training.validation_metrics.multiscale.loss_matrices_path = str(Path(tmp_path_loss_matrices).parent)
    return cfg


@pytest.fixture
def ensemble_config(
    testing_modifications_callbacks_on_with_temp_dir: DictConfig,
    get_tmp_paths: GetTmpPaths,
    get_test_data: GetTestData,
) -> tuple[DictConfig, str]:
    overrides = ["model=graphtransformer_ens", "graph=multi_scale"]

    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_ensemble_crps"):
        template = compose(config_name="ensemble_crps", overrides=overrides)

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_ensemble_crps.yaml")
    assert isinstance(use_case_modifications, DictConfig)

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.system.input.dataset = str(Path(tmp_dir, rel_paths[0]))

    cfg = OmegaConf.merge(template, testing_modifications_callbacks_on_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)

    cfg = handle_truncation_matrices(cfg, get_test_data)

    assert isinstance(cfg, DictConfig)
    return cfg, dataset_urls[0]


@pytest.fixture
def hierarchical_config(
    testing_modifications_callbacks_on_with_temp_dir: DictConfig,
    get_tmp_paths: GetTmpPaths,
) -> tuple[DictConfig, list[str]]:
    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_hierarchical"):
        template = compose(config_name="hierarchical")

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_config.yaml")
    assert isinstance(use_case_modifications, DictConfig)

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.system.input.dataset = str(Path(tmp_dir, rel_paths[0]))

    cfg = OmegaConf.merge(template, testing_modifications_callbacks_on_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    assert isinstance(cfg, DictConfig)
    return cfg, dataset_urls


@pytest.fixture
def gnn_config(testing_modifications_with_temp_dir: DictConfig, get_tmp_paths: GetTmpPaths) -> tuple[DictConfig, str]:
    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_config"):
        template = compose(config_name="config")

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_config.yaml")
    assert isinstance(use_case_modifications, DictConfig)

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.system.input.dataset = str(Path(tmp_dir, rel_paths[0]))

    cfg = OmegaConf.merge(template, testing_modifications_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    assert isinstance(cfg, DictConfig)
    return cfg, dataset_urls[0]


@pytest.fixture(
    params=[  # selects different test cases
        "lam",
        "graphtransformer",
        "stretched",
        "ensemble_crps",
    ],
)
def benchmark_config(
    request: pytest.FixtureRequest,
    testing_modifications_with_temp_dir: OmegaConf,
    get_test_data: GetTestData,
) -> tuple[OmegaConf, str]:
    test_case = request.param
    base_config = "config"  # which config we start from in anemoi/training/configs/
    # base_config="config" =>  anemoi/training/configs/config.yaml
    # LAM and Stretched need different base configs

    # change configs based on test case
    if test_case == "graphtransformer":
        overrides = ["model=graphtransformer", "graph=multi_scale"]
    elif test_case == "stretched":
        overrides = []
        base_config = "stretched"
    elif test_case == "lam":
        overrides = []
        base_config = "lam"
    elif test_case == "ensemble_crps":
        overrides = ["model=graphtransformer_ens", "graph=multi_scale"]
        base_config = "ensemble_crps"
    else:
        msg = f"Error. Unknown benchmark configuration: {test_case}"
        raise ValueError(msg)

    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="benchmark"):
        template = compose(config_name=base_config, overrides=overrides)

    # Settings for benchmarking in general (sets atos paths, enables profiling, disables plotting etc)
    base_benchmark_config = OmegaConf.load(Path.cwd() / Path("training/tests/integration/config/benchmark/base.yaml"))
    # Settings for the specific benchmark test case
    use_case_modifications = OmegaConf.load(
        Path.cwd() / f"training/tests/integration/config/benchmark/{test_case}.yaml",
    )
    cfg = OmegaConf.merge(template, testing_modifications_with_temp_dir, use_case_modifications, base_benchmark_config)

    cfg.system.output.profiler = Path(cfg.system.output.root + "/" + cfg.system.output.profiler)
    OmegaConf.resolve(cfg)

    if test_case == "ensemble_crps":
        cfg = handle_truncation_matrices(cfg, get_test_data)
    return cfg, test_case


@pytest.fixture(scope="session")
def migrator() -> Migrator:
    return Migrator()


@pytest.fixture(
    params=[
        ["model=gnn"],
        ["model=graphtransformer"],
    ],
)
def architecture_config_with_checkpoint(
    migrator: Migrator,
    request: pytest.FixtureRequest,
    testing_modifications_with_temp_dir: DictConfig,
    get_tmp_paths: GetTmpPaths,
    get_test_data: GetTestData,
) -> tuple[OmegaConf, str]:
    # Reuse the same overrides that architecture_config gets
    overrides = request.param
    cfg, dataset_url, model_architecture = build_architecture_config(
        overrides,
        testing_modifications_with_temp_dir,
        get_tmp_paths,
    )
    # rest of your logic...
    if "gnn" in model_architecture:
        existing_ckpt = get_test_data(
            "anemoi-integration-tests/training/checkpoints/testing-checkpoint-gnn-global-2025-07-31.ckpt",
        )
    elif "graphtransformer" in model_architecture:
        existing_ckpt = get_test_data(
            "anemoi-integration-tests/training/checkpoints/testing-checkpoint-graphtransformer-global-2025-07-31.ckpt",
        )
    else:
        msg = f"Unknown architecture in config {cfg.model.architecture}"
        raise ValueError(msg)

    _, new_ckpt, _ = migrator.sync(existing_ckpt)

    checkpoint_dir = Path(cfg.system.output.root + "/" + cfg.system.output.checkpoints.root + "/dummy_id")
    checkpoint_dir.mkdir(parents=True, exist_ok=True)
    torch.save(new_ckpt, checkpoint_dir / "last.ckpt")

    cfg.training.run_id = "dummy_id"
    cfg.training.max_epochs = 3

    return cfg, dataset_url


@pytest.fixture
def interpolator_config(
    testing_modifications_callbacks_on_with_temp_dir: DictConfig,
    get_tmp_paths: GetTmpPaths,
) -> tuple[DictConfig, str]:
    """Compose a runnable configuration for the temporal-interpolation model.

    It is based on `interpolator.yaml` and only patches paths pointing to the
    sample dataset that the tests download locally.
    """
    # No model override here - the template already sets the dedicated
    # interpolator model + GraphInterpolator Lightning task.
    with initialize(
        version_base=None,
        config_path="../../src/anemoi/training/config",
        job_name="test_interpolator",
    ):
        template = compose(config_name="interpolator")

    use_case_modifications = OmegaConf.load(
        Path.cwd() / "training/tests/integration/config/test_interpolator.yaml",
    )
    assert isinstance(use_case_modifications, DictConfig)

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.system.input.dataset = str(Path(tmp_dir, rel_paths[0]))
    cfg = OmegaConf.merge(template, testing_modifications_callbacks_on_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    assert isinstance(cfg, DictConfig)
    return cfg, dataset_urls[0]


@pytest.fixture(
    params=[
        [],
        [
            "model=graphtransformer_diffusiontend",
            "training.model_task=anemoi.training.train.tasks.GraphDiffusionTendForecaster",
        ],
    ],
    ids=["diffusion", "diffusiontend"],
)
def diffusion_config(
    request: pytest.FixtureRequest,
    testing_modifications_callbacks_on_with_temp_dir: OmegaConf,
    get_tmp_paths: callable,
) -> tuple[OmegaConf, str]:
    overrides = request.param

    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_diffusion"):
        template = compose(config_name="diffusion", overrides=overrides)

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_diffusion.yaml")
    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.system.input.dataset = str(Path(tmp_dir, rel_paths[0]))

    cfg = OmegaConf.merge(template, testing_modifications_callbacks_on_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    return cfg, dataset_urls[0]



# (C) Copyright 2024- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from __future__ import annotations

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from anemoi.models.schemas.data_processor import PreprocessorSchema  # noqa: TC002


class DataSchema(PydanticBaseModel):
    """A class used to represent the overall configuration of the dataset.

    Attributes
    ----------
    format : str
        The format of the data.
    resolution : str
        The resolution of the data.
    frequency : str
        The frequency of the data.
    timestep : str
        The timestep of the data.
    forcing : list[str]
        The list of features used as forcing to generate the forecast state.
    diagnostic : list[str]
        The list of features that are only part of the forecast state.
    processors : Dict[str, Processor]
        The Processors configuration.
    target : list[str], optional
        The list of features used to compute the loss against forecasted variables.
    num_features : int, optional
        The number of features in the forecast state. To be set in the code.
    """

    format: str = Field(example=None)
    "Format of the data."
    frequency: str = Field(example=None)
    "Time frequency requested from the dataset."
    timestep: str = Field(example=None)
    "Time step of model (must be multiple of frequency)."
    processors: dict[str, PreprocessorSchema]
    "Layers of model performing computation on latent space. \
            Processors including imputers and normalizers are applied in order of definition."
    forcing: list[str] = Field(default_factory=list)
    "Features that are not part of the forecast state but are used as forcing to generate the forecast state."
    diagnostic: list[str] = Field(default_factory=list)
    "Features that are only part of the forecast state and are not used as an input to the model."
    target: list[str] | None = None
    (
        "Features used to compute the loss against forecasted variables. "
        "Cannot be prognostic or diagnostic, can have the same name as forcing variables "
        "but have a different role. Such that: prognostic = diagnostic - forcing.union(target)."
    )
    num_features: int | None
    "Number of features in the forecast state. To be set in the code."



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from collections.abc import Callable
from functools import cached_property

import numpy as np
import pytorch_lightning as pl
from hydra.utils import instantiate
from torch.utils.data import DataLoader
from torch_geometric.data import HeteroData

from anemoi.datasets import open_dataset
from anemoi.models.data_indices.collection import IndexCollection
from anemoi.training.data.dataset import NativeGridDataset
from anemoi.training.data.grid_indices import BaseGridIndices
from anemoi.training.schemas.base_schema import BaseSchema
from anemoi.training.utils.worker_init import worker_init_func
from anemoi.utils.dates import frequency_to_seconds

LOGGER = logging.getLogger(__name__)


class AnemoiDatasetsDataModule(pl.LightningDataModule):
    """Anemoi Datasets data module for PyTorch Lightning."""

    def __init__(self, config: BaseSchema, graph_data: HeteroData) -> None:
        """Initialize Anemoi Datasets data module.

        Parameters
        ----------
        config : BaseSchema
            Job configuration

        """
        super().__init__()

        self.config = config
        self.graph_data = graph_data

        # Set the training end date if not specified
        if self.config.dataloader.training.end is None:
            LOGGER.info(
                "No end date specified for training data, setting default before validation start date %s.",
                self.config.dataloader.validation.start - 1,
            )
            self.config.dataloader.training.end = self.config.dataloader.validation.start - 1

        if not self.config.dataloader.pin_memory:
            LOGGER.info("Data loader memory pinning disabled.")

    @cached_property
    def statistics(self) -> dict:
        return self.ds_train.statistics

    @cached_property
    def statistics_tendencies(self) -> dict:
        return self.ds_train.statistics_tendencies

    @cached_property
    def metadata(self) -> dict:
        return self.ds_train.metadata

    @cached_property
    def supporting_arrays(self) -> dict:
        return self.ds_train.supporting_arrays | self.grid_indices.supporting_arrays

    @cached_property
    def data_indices(self) -> IndexCollection:
        return IndexCollection(self.config, self.ds_train.name_to_index)

    def relative_date_indices(self, val_rollout: int = 1) -> list:
        """Determine a list of relative time indices to load for each batch."""
        if hasattr(self.config.training, "explicit_times"):
            return sorted(set(self.config.training.explicit_times.input + self.config.training.explicit_times.target))

        # Calculate indices using multistep, timeincrement and rollout.
        # Use the maximum rollout to be expected
        rollout_cfg = getattr(getattr(self.config, "training", None), "rollout", None)

        rollout_max = getattr(rollout_cfg, "max", None)
        rollout_start = getattr(rollout_cfg, "start", 1)
        rollout_epoch_increment = getattr(rollout_cfg, "epoch_increment", 0)

        # Fallback if max is None or rollout_cfg is missing
        rollout_value = rollout_start
        if rollout_cfg and rollout_epoch_increment > 0 and rollout_max is not None:
            rollout_value = rollout_max

        else:
            LOGGER.warning(
                "Falling back rollout to: %s",
                rollout_value,
            )

        rollout = max(rollout_value, val_rollout)

        multi_step = self.config.training.multistep_input
        return [self.timeincrement * mstep for mstep in range(multi_step + rollout)]

    def add_trajectory_ids(self, data_reader: Callable) -> Callable:
        """Determine an index of forecast trajectories associated with the time index and add to a data_reader object.

        This is needed for interpolation to ensure that the interpolator is trained on consistent time slices.

        NOTE: This is only relevant when training on non-analysis and could in the future be replaced with
        a property of the dataset stored in data_reader. Now assumes regular interval of changed model runs
        """
        if not hasattr(self.config.dataloader, "model_run_info"):
            data_reader.trajectory_ids = None
            return data_reader

        mr_start = np.datetime64(self.config.dataloader.model_run_info.start)
        mr_len = self.config.dataloader.model_run_info.length  # model run length in number of date indices
        if hasattr(self.config.training, "rollout") and self.config.training.rollout.max is not None:
            max_rollout_index = max(self.relative_date_indices(self.config.training.rollout.max))
            assert (
                max_rollout_index < mr_len
            ), f"""Requested data length {max_rollout_index + 1}
                    longer than model run length {mr_len}"""

        data_reader.trajectory_ids = (data_reader.dates - mr_start) // np.timedelta64(
            mr_len * frequency_to_seconds(self.config.data.frequency),
            "s",
        )
        return data_reader

    @cached_property
    def grid_indices(self) -> type[BaseGridIndices]:
        reader_group_size = self.config.dataloader.read_group_size

        grid_indices = instantiate(
            self.config.dataloader.grid_indices,
            reader_group_size=reader_group_size,
        )
        grid_indices.setup(self.graph_data)
        return grid_indices

    @cached_property
    def timeincrement(self) -> int:
        """Determine the step size relative to the data frequency."""
        try:
            frequency = frequency_to_seconds(self.config.data.frequency)
        except ValueError as e:
            msg = f"Error in data frequency, {self.config.data.frequency}"
            raise ValueError(msg) from e

        try:
            timestep = frequency_to_seconds(self.config.data.timestep)
        except ValueError as e:
            msg = f"Error in timestep, {self.config.data.timestep}"
            raise ValueError(msg) from e

        assert timestep % frequency == 0, (
            f"Timestep ({self.config.data.timestep} == {timestep}) isn't a "
            f"multiple of data frequency ({self.config.data.frequency} == {frequency})."
        )

        LOGGER.info(
            "Timeincrement set to %s for data with frequency, %s, and timestep, %s",
            timestep // frequency,
            frequency,
            timestep,
        )
        return timestep // frequency

    @cached_property
    def ds_train(self) -> NativeGridDataset:
        return self._get_dataset(
            open_dataset(self.config.dataloader.training),
            label="train",
        )

    @cached_property
    def ds_valid(self) -> NativeGridDataset:
        if not self.config.dataloader.training.end < self.config.dataloader.validation.start:
            LOGGER.warning(
                "Training end date %s is not before validation start date %s.",
                self.config.dataloader.training.end,
                self.config.dataloader.validation.start,
            )
        return self._get_dataset(
            open_dataset(self.config.dataloader.validation),
            shuffle=False,
            val_rollout=self.config.dataloader.validation_rollout,
            label="validation",
        )

    @cached_property
    def ds_test(self) -> NativeGridDataset:
        assert self.config.dataloader.training.end < self.config.dataloader.test.start, (
            f"Training end date {self.config.dataloader.training.end} is not before"
            f"test start date {self.config.dataloader.test.start}"
        )
        assert self.config.dataloader.validation.end < self.config.dataloader.test.start, (
            f"Validation end date {self.config.dataloader.validation.end} is not before"
            f"test start date {self.config.dataloader.test.start}"
        )
        return self._get_dataset(
            open_dataset(self.config.dataloader.test),
            shuffle=False,
            label="test",
        )

    def _get_dataset(
        self,
        data_reader: Callable,
        shuffle: bool = True,
        val_rollout: int = 1,
        label: str = "generic",
    ) -> NativeGridDataset:

        data_reader = self.add_trajectory_ids(data_reader)  # NOTE: Functionality to be moved to anemoi datasets

        return NativeGridDataset(
            data_reader=data_reader,
            relative_date_indices=self.relative_date_indices(val_rollout),
            timestep=self.config.data.timestep,
            shuffle=shuffle,
            grid_indices=self.grid_indices,
            label=label,
        )

    def _get_dataloader(self, ds: NativeGridDataset, stage: str) -> DataLoader:
        assert stage in {"training", "validation", "test"}
        return DataLoader(
            ds,
            batch_size=self.config.dataloader.batch_size[stage],
            # number of worker processes
            num_workers=self.config.dataloader.num_workers[stage],
            # use of pinned memory can speed up CPU-to-GPU data transfers
            # see https://pytorch.org/docs/stable/notes/cuda.html#cuda-memory-pinning
            pin_memory=self.config.dataloader.pin_memory,
            # worker initializer
            worker_init_fn=worker_init_func,
            # prefetch batches
            prefetch_factor=self.config.dataloader.prefetch_factor,
            persistent_workers=True,
        )

    def train_dataloader(self) -> DataLoader:
        return self._get_dataloader(self.ds_train, "training")

    def val_dataloader(self) -> DataLoader:
        return self._get_dataloader(self.ds_valid, "validation")

    def test_dataloader(self) -> DataLoader:
        return self._get_dataloader(self.ds_test, "test")



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
# Various tests of the Anemoi components using a sample data set.
#
# This script is not part of a productive ML workflow, but is
# used for CI/CD!
import os
import pathlib
from functools import reduce
from operator import getitem
from pathlib import Path

import matplotlib as mpl
import pytest
import torch
from conftest import GetTmpPaths
from hydra import compose
from hydra import initialize
from omegaconf import DictConfig
from typeguard import typechecked

import anemoi.training
from anemoi.training.schemas.base_schema import BaseSchema
from anemoi.training.train.train import AnemoiTrainer
from anemoi.utils.testing import GetTestArchive
from anemoi.utils.testing import GetTestData

os.environ["ANEMOI_BASE_SEED"] = "42"
os.environ["ANEMOI_CONFIG_PATH"] = str(pathlib.Path(anemoi.training.__file__).parent / "config")
mpl.use("agg")


@pytest.fixture
@typechecked
def aicon_config_with_tmp_dir(get_tmp_paths: GetTmpPaths, get_test_archive: GetTestArchive) -> DictConfig:
    """Get AICON config with temporary output paths."""
    with initialize(version_base=None, config_path="./"):
        config = compose(config_name="test_cicd_aicon_04_icon-dream_medium")

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(config, ["dataset", "forcing_dataset"])
    config.system.output.root = tmp_dir
    dataset, forcing_dataset = rel_paths
    config.system.input.dataset = str(Path(tmp_dir, dataset))
    config.system.input.forcing_dataset = str(Path(tmp_dir, forcing_dataset))

    for url in dataset_urls:
        get_test_archive(url)

    return config


@pytest.fixture
@typechecked
def aicon_config_with_grid(aicon_config_with_tmp_dir: DictConfig, get_test_data: GetTestData) -> DictConfig:
    """Temporarily download the ICON grid specified in the config.

    Downloading the grid is required as the AICON grid is currently required as a netCDF file.
    """
    aicon_config_with_tmp_dir.graph.nodes.icon_mesh.node_builder.grid_filename = get_test_data(
        aicon_config_with_tmp_dir.graph.nodes.icon_mesh.node_builder.grid_filename,
    )
    return aicon_config_with_tmp_dir


@pytest.fixture
@typechecked
def trained_aicon(aicon_config_with_grid: DictConfig) -> tuple[AnemoiTrainer, float, float]:
    """Train AICON and return testable objects."""
    trainer = AnemoiTrainer(aicon_config_with_grid)
    initial_sum = float(torch.tensor(list(map(torch.sum, trainer.model.parameters()))).sum())
    trainer.train()
    final_sum = float(torch.tensor(list(map(torch.sum, trainer.model.parameters()))).sum())
    return trainer, initial_sum, final_sum


@typechecked
def assert_metadatakeys(metadata: dict, *metadata_keys: tuple[str, ...]) -> None:
    """Assert the presence of all `metadata_keys` in `metadata`.

    A metadata key is a tuple that describes a path within the `metadata` dict of dicts.
    E.g., if `metadata_keys[0] = ("a", "b", "c")`, this test will verify the existence of `metadata["a"]["b"]["c"]`.
    """
    errors = []
    for keys in metadata_keys:
        try:
            reduce(getitem, keys, metadata)
        except KeyError:  # noqa: PERF203
            keys = "".join(f"[{k!r}]" for k in keys)
            errors.append("missing metadata" + keys)
    if errors:
        raise KeyError("\n".join(errors))


@typechecked
def test_config_validation_aicon(aicon_config_with_tmp_dir: DictConfig) -> None:
    BaseSchema(**aicon_config_with_tmp_dir)


@pytest.mark.slow
@typechecked
def test_aicon_metadata(aicon_config_with_grid: DictConfig) -> None:
    """Test for presence of metadata required for inference.

    The objective of this test is to monitor changes to the metadata structure and ensure compatibility.
    Please update the path to each datum accordingly whenever a revision to the metadata structure is implemented.
    """
    trainer = AnemoiTrainer(aicon_config_with_grid)

    assert_metadatakeys(
        trainer.metadata,
        ("config", "data", "timestep"),
        ("config", "graph", "nodes", "icon_mesh", "node_builder", "max_level_dataset"),
        ("config", "training", "precision"),
        ("data_indices", "data", "input", "diagnostic"),
        ("data_indices", "data", "input", "full"),
        ("data_indices", "data", "output", "full"),
        ("data_indices", "model", "input", "forcing"),
        ("data_indices", "model", "input", "full"),
        ("data_indices", "model", "input", "prognostic"),
        ("data_indices", "model", "output", "full"),
        ("dataset", "shape"),
    )

    assert torch.is_tensor(trainer.graph_data["data"].x), "data coordinates not present"

    # Assert heterogeneity of num_chunks setting.
    assert aicon_config_with_grid.model.encoder.num_chunks != aicon_config_with_grid.model.decoder.num_chunks

    # Monitor path and setting of num_chunks
    assert trainer.model.model.model.encoder.proc.num_chunks == aicon_config_with_grid.model.encoder.num_chunks
    assert trainer.model.model.model.decoder.proc.num_chunks == aicon_config_with_grid.model.decoder.num_chunks


@pytest.mark.slow
@typechecked
def test_aicon_training(trained_aicon: tuple) -> None:
    _, initial_sum, final_sum = trained_aicon
    assert initial_sum != final_sum



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import pytorch_lightning as pl
import torch
from omegaconf import OmegaConf
from pytorch_lightning.callbacks import Callback

LOGGER = logging.getLogger(__name__)


class ParentUUIDCallback(Callback):
    """A callback that retrieves the parent UUID for a model, if it is a child model."""

    def __init__(self, config: OmegaConf) -> None:
        """Initialise the ParentUUIDCallback callback.

        Parameters
        ----------
        config : OmegaConf
            Config object

        """
        super().__init__()
        self.config = config

    def on_load_checkpoint(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        checkpoint: torch.nn.Module,
    ) -> None:
        del trainer  # unused
        pl_module.hparams["metadata"]["parent_uuid"] = checkpoint["hyper_parameters"]["metadata"]["uuid"]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import shutil
from pathlib import Path

import pytest
import torch
from omegaconf import DictConfig
from omegaconf import OmegaConf
from pytorch_lightning import Trainer
from pytorch_lightning.demos.boring_classes import BoringModel
from torch import nn

from anemoi.training.diagnostics.callbacks import AnemoiCheckpoint
from anemoi.training.utils.jsonify import map_config_to_primitives
from anemoi.utils.checkpoints import load_metadata
from anemoi.utils.config import DotDict


class DummyModel(torch.nn.Module):
    """Dummy pytorch model for testing."""

    def __init__(self, *, config: DotDict, metadata: dict):
        super().__init__()

        self.config = config
        self.metadata = metadata
        self.supporting_arrays = {}
        self.fc1 = nn.Linear(32, 5)
        self.fc2 = nn.Linear(5, 1)
        self.relu = nn.ReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.relu(x)
        return self.fc2(x)


class DummyModule(BoringModel):
    """Dummy lightning module for testing."""

    def __init__(self, *, config: DotDict, metadata: dict) -> None:
        super().__init__()
        self.model = DummyModel(config=config, metadata=metadata)
        self.save_hyperparameters()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


@pytest.fixture
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> str:
    # returns a session-scoped temporary directory
    return str(tmp_path_factory.mktemp("test_checkpoint"))


@pytest.fixture
def config() -> DictConfig:
    config = DictConfig({"diagnostics": {"log": {"wandb": {"enabled": False}, "mlflow": {"enabled": False}}}})
    return DotDict(map_config_to_primitives(OmegaConf.to_container(config, resolve=True)))


@pytest.fixture
def checkpoint_settings(tmp_path: str) -> dict[str : str | bool]:
    return {
        "dirpath": tmp_path,
        "verbose": False,
        "save_weights_only": False,
        "auto_insert_metric_name": False,
        "save_on_train_epoch_end": False,
        "enable_version_counter": False,
    }


@pytest.fixture
def callback(tmp_path: str, config: DictConfig, checkpoint_settings: dict) -> AnemoiCheckpoint:
    callback = AnemoiCheckpoint(
        config=config,
        filename="{step}",
        save_last=True,
        train_time_interval=datetime.timedelta(seconds=0.1),
        save_top_k=3,
        monitor="step",
        mode="max",
        **checkpoint_settings,
    )
    callback.dirpath = tmp_path
    return callback


@pytest.fixture
def metadata(config: DictConfig) -> dict:
    return map_config_to_primitives(
        {
            "config": config,
        },
    )


@pytest.fixture
def model(metadata: dict, config: DictConfig) -> DummyModule:
    kwargs = {
        "metadata": metadata,
        "config": config,
    }
    return DummyModule(**kwargs)


@pytest.mark.data_dependent
def test_same_uuid(tmp_path: str, callback: AnemoiCheckpoint, model: DummyModule) -> None:
    """Test if the inference checkpoints and lightning checkpoints store same uuid.

    Args:
        tmp_path (str): path to checkpoint dir
        callback (AnemoiCheckpoint): callback to store checkpoints
        model (DummyModule): dummy lightning module
    """
    trainer = Trainer(default_root_dir=tmp_path, accelerator="cpu", callbacks=[callback], max_epochs=2, logger=False)
    trainer.fit(model)

    for ckpt_path in Path(tmp_path).iterdir():
        if str(ckpt_path.name).startswith("inference"):
            pl_ckpt_name = (ckpt_path.name).replace("inference-", "")

            if Path(tmp_path + "/" + pl_ckpt_name).exists():
                uuid = load_metadata(ckpt_path)["uuid"]

                pl_model = DummyModule.load_from_checkpoint(tmp_path + "/" + pl_ckpt_name, weights_only=False)

                assert uuid == pl_model.hparams["metadata"]["uuid"]

    shutil.rmtree(tmp_path)
